import os
import os.path
import re
import time
from collections import OrderedDict
from operator import attrgetter
from threading import Lock
from typing import TYPE_CHECKING

import dateutil.parser

import sickchill
from sickchill import logger
from sickchill.helper.common import remove_extension
from sickchill.oldbeard import common, db, helpers, scene_exceptions, scene_numbering
from sickchill.oldbeard.name_parser import regexes

if TYPE_CHECKING:
    from sickchill.tv import TVShow


class NameParser(object):
    ALL_REGEX = 0
    NORMAL_REGEX = 1
    ANIME_REGEX = 2

    def __init__(self, filename=True, showObj=None, tryIndexers=False, naming_pattern=False, parse_method=None):

        self.filename = filename
        self.showObj = showObj
        self.tryIndexers = tryIndexers

        self.naming_pattern = naming_pattern

        if (self.showObj and not self.showObj.is_anime) or parse_method == "normal":
            self._compile_regexes(self.NORMAL_REGEX)
        elif (self.showObj and self.showObj.is_anime) or parse_method == "anime":
            self._compile_regexes(self.ANIME_REGEX)
        else:
            self._compile_regexes(self.ALL_REGEX)

    @staticmethod
    def clean_series_name(series_name):
        """Cleans up series name by removing any . and _
        characters, along with any trailing hyphens.

        Is basically equivalent to replacing all _ and . with a
        space, but handles decimal numbers in string.
        Stolen from dbr's tvnamer
        """

        series_name = re.sub(r"(\D)\.(?!\s)(\D)", "\\1 \\2", series_name)
        series_name = re.sub(r"(\d)\.(\d{4})", "\\1 \\2", series_name)  # if it ends in a year then don't keep the dot
        series_name = re.sub(r"(\D)\.(?!\s)", "\\1 ", series_name)
        series_name = re.sub(r"\.(?!\s)(\D)", " \\1", series_name)
        series_name = series_name.replace("_", " ")
        series_name = re.sub(r"-$", "", series_name)
        series_name = re.sub(r"^\[.*\]", "", series_name)
        return series_name.strip()

    def _compile_regexes(self, regexMode):
        if regexMode == self.ANIME_REGEX:
            dbg_str = "ANIME"
            uncompiled_regex = [regexes.anime_regexes]
        elif regexMode == self.NORMAL_REGEX:
            dbg_str = "NORMAL"
            uncompiled_regex = [regexes.normal_regexes]
        else:
            dbg_str = "ALL"
            uncompiled_regex = [regexes.normal_regexes, regexes.anime_regexes]

        self.compiled_regexes = []
        for regexItem in uncompiled_regex:
            for cur_pattern_num, (cur_pattern_name, cur_pattern) in enumerate(regexItem):
                try:
                    cur_regex = re.compile(cur_pattern, re.VERBOSE | re.I)
                except re.error as errormsg:
                    logger.info(f"WARNING: Invalid episode_pattern using {dbg_str} regexs, {errormsg}. {cur_pattern}")
                else:
                    self.compiled_regexes.append((cur_pattern_num, cur_pattern_name, cur_regex))

    def _parse_string(self, name, skip_scene_detection=False):
        if not name:
            return

        matches = []
        best_result = None

        for (cur_regex_num, cur_regex_name, cur_regex) in self.compiled_regexes:
            match = cur_regex.match(name)

            if not match:
                continue

            result = ParseResult(name)
            result.which_regex = [cur_regex_name]
            result.score = 0 - cur_regex_num

            named_groups = list(match.groupdict())

            if "series_name" in named_groups:
                result.series_name = match.group("series_name")
                if result.series_name:
                    result.series_name = self.clean_series_name(result.series_name)
                    result.score += 1

            if "series_num" in named_groups and match.group("series_num"):
                result.score += 1

            if "season_num" in named_groups:
                tmp_season = int(match.group("season_num"))
                if cur_regex_name == "bare" and tmp_season in (19, 20):
                    continue
                if cur_regex_name == "fov" and tmp_season > 500:
                    continue

                result.season_number = tmp_season
                result.score += 1

            if "ep_num" in named_groups:
                ep_num = self._convert_number(match.group("ep_num"))
                if "extra_ep_num" in named_groups and match.group("extra_ep_num"):
                    tmp_episodes = list(range(ep_num, self._convert_number(match.group("extra_ep_num")) + 1))
                    if len(tmp_episodes) > 4:
                        continue
                else:
                    tmp_episodes = [ep_num]

                result.episode_numbers = tmp_episodes
                result.score += 3

            if "ep_ab_num" in named_groups:
                ep_ab_num = self._convert_number(match.group("ep_ab_num"))
                if "extra_ab_ep_num" in named_groups and match.group("extra_ab_ep_num"):
                    result.ab_episode_numbers = list(range(ep_ab_num, self._convert_number(match.group("extra_ab_ep_num")) + 1))
                    result.score += 1
                else:
                    result.ab_episode_numbers = [ep_ab_num]
                result.score += 1

            if "air_date" in named_groups:
                air_date = match.group("air_date")
                try:
                    # Workaround for shows that get interpreted as 'air_date' incorrectly.
                    # Shows so far are 11.22.63 and 9-1-1
                    excluded_shows = ["112263", "911"]
                    assert re.sub(r"[^\d]*", "", air_date) not in excluded_shows

                    # noinspection PyUnresolvedReferences
                    check = dateutil.parser.parse(air_date, fuzzy_with_tokens=True)[0].date()
                    # Make sure a 20th century date isn't returned as a 21st century date
                    # 1 Year into the future (No releases should be coming out a year ahead of time, that's just insane)
                    if check > check.today() and (check - check.today()).days // 365 > 1:
                        check = check.replace(year=check.year - 100)

                    result.air_date = check
                    result.score += 1
                except Exception as error:
                    logger.debug(error)
                    continue

            if "extra_info" in named_groups:
                tmp_extra_info = match.group("extra_info")

                # Show.S04.Special or Show.S05.Part.2.Extras is almost certainly not every episode in the season
                if tmp_extra_info and cur_regex_name == "season_only" and re.search(r"([. _-]|^)(special|extra)s?\w*([. _-]|$)", tmp_extra_info, re.I):
                    continue
                result.extra_info = tmp_extra_info
                result.score += 1

            if "release_group" in named_groups:
                result.release_group = match.group("release_group")
                result.score += 1

            if "version" in named_groups:
                # assigns version to anime file if detected using anime regex. Non-anime regex receives -1
                version = match.group("version")
                if version:
                    result.version = version
                else:
                    result.version = 1
            else:
                result.version = -1

            matches.append(result)

        # only get matches with series_name
        # TODO: This makes tests fail when checking filenames that do not include the show name (refresh, force update, etc)
        # matches = [x for x in matches if x.series_name]

        if matches:
            # pick best match with highest score based on placement
            best_result = max(sorted(matches, reverse=True, key=attrgetter("which_regex")), key=attrgetter("score"))

            show = None
            if best_result and best_result.series_name and not self.naming_pattern:
                # try and create a show object for this result
                show = helpers.get_show(best_result.series_name, self.tryIndexers)

            # confirm passed in show object indexer id matches result show object indexer id
            if show:
                if self.showObj and show.indexerid != self.showObj.indexerid:
                    show = None
                best_result.show = show
            elif self.showObj and not show:
                best_result.show = self.showObj

            # Only allow anime matches if resolved show or specified show is anime
            best_result = self.check_anime_preferred(best_result, matches)

            # if this is a naming pattern test or result doesn't have a show object then return best result
            if not best_result.show or self.naming_pattern:
                return best_result

            # get quality
            best_result.quality = common.Quality.nameQuality(name, best_result.show.is_anime)

            new_episode_numbers = []
            new_season_numbers = []
            new_absolute_numbers = []

            # if we have an air-by-date show then get the real season/episode numbers
            if best_result.is_air_by_date:
                airdate = best_result.air_date.toordinal()
                main_db_con = db.DBConnection()
                sql_result = main_db_con.select(
                    "SELECT season, episode FROM tv_episodes WHERE showid = ? and indexer = ? and airdate = ?",
                    [best_result.show.indexerid, best_result.show.indexer, airdate],
                )

                season_number = None
                episode_numbers = []

                if sql_result:
                    season_number = int(sql_result[0][0])
                    episode_numbers = [int(sql_result[0][1])]

                if season_number is None or not episode_numbers:
                    try:
                        epObj = sickchill.indexer.episode(best_result.show, firstAired=best_result.air_date)
                        season_number = epObj["airedSeason"]
                        episode_numbers = [epObj["airedEpisode"]]
                    except Exception:
                        logger.warning(f"Unable to find episode with date {best_result.air_date} for show {best_result.show.name}, skipping")
                        episode_numbers = []

                for epNo in episode_numbers:
                    s = season_number
                    e = epNo

                    if best_result.show.is_scene:
                        (s, e) = scene_numbering.get_indexer_numbering(best_result.show.indexerid, best_result.show.indexer, season_number, epNo)
                    new_episode_numbers.append(e)
                    new_season_numbers.append(s)

            elif best_result.show.is_anime and best_result.ab_episode_numbers:
                best_result.scene_season = scene_exceptions.get_scene_exception_by_name(best_result.series_name)[1]
                for epAbsNo in best_result.ab_episode_numbers:
                    a = epAbsNo

                    if best_result.show.is_scene and not skip_scene_detection:
                        a = scene_numbering.get_indexer_absolute_numbering(
                            best_result.show.indexerid, best_result.show.indexer, epAbsNo, True, best_result.scene_season
                        )

                    (s, e) = helpers.get_all_episodes_from_absolute_number(best_result.show, [a])

                    new_absolute_numbers.append(a)
                    new_episode_numbers.extend(e)
                    new_season_numbers.append(s)

            elif best_result.season_number and best_result.episode_numbers:
                for epNo in best_result.episode_numbers:
                    s = best_result.season_number
                    e = epNo

                    if best_result.show.is_scene and not skip_scene_detection:
                        (s, e) = scene_numbering.get_indexer_numbering(best_result.show.indexerid, best_result.show.indexer, best_result.season_number, epNo)
                    if best_result.show.is_anime:
                        a = helpers.get_absolute_number_from_season_and_episode(best_result.show, s, e)
                        if a:
                            new_absolute_numbers.append(a)

                    new_episode_numbers.append(e)
                    new_season_numbers.append(s)

            # need to do a quick sanity check heregex.  It's possible that we now have episodes
            # from more than one season (by tvdb numbering), and this is just too much
            # for oldbeard, so we'd need to flag it.
            new_season_numbers = list(set(new_season_numbers))  # remove duplicates
            if len(new_season_numbers) > 1:
                raise InvalidNameException(
                    f"Scene numbering results episodes from seasons {new_season_numbers}, " f"(i.e. more than one) and sickchill does not support this. Sorry."
                )

            # I guess it's possible that we'd have duplicate episodes too, so lets
            # eliminate them
            new_episode_numbers = sorted(set(new_episode_numbers))

            # maybe even duplicate absolute numbers so why not do them as well
            new_absolute_numbers = list(set(new_absolute_numbers))
            new_absolute_numbers.sort()

            if new_absolute_numbers:
                best_result.ab_episode_numbers = new_absolute_numbers

            if new_season_numbers and new_episode_numbers:
                best_result.episode_numbers = new_episode_numbers
                best_result.season_number = new_season_numbers[0]

            if best_result.show.is_scene and not skip_scene_detection:
                logger.debug(f"Converted parsed result {best_result.original_name} into {best_result}")

        # CPU sleep
        time.sleep(0.02)

        return best_result

    def check_anime_preferred(self, best_result, matches):
        show = self.showObj or best_result.show
        if (best_result.show and best_result.show.is_anime and not self.showObj) or (self.showObj and self.showObj.is_anime):
            anime_matches = [x for x in matches if x.is_anime]
            if anime_matches:
                best_result_anime = max(sorted(anime_matches, reverse=True, key=attrgetter("which_regex")), key=attrgetter("score"))
                if best_result_anime and best_result_anime.series_name:
                    show_anime = helpers.get_show(best_result_anime.series_name)
                    if show_anime and show_anime.indexerid == show.indexerid:
                        best_result = best_result_anime

        return best_result

    @staticmethod
    def _combine_results(first, second, attr):
        # if the first doesn't exist then return the second or nothing
        if not first:
            if not second:
                return None
            else:
                return getattr(second, attr)

        # if the second doesn't exist then return the first
        if not second:
            return getattr(first, attr)

        a = getattr(first, attr)
        b = getattr(second, attr)

        # if a is good use it
        if a is not None or (isinstance(a, list) and a):
            return a
        # if not use b (if b isn't set it'll just be default)
        else:
            return b

    @staticmethod
    def _unicodify(obj, encoding="utf-8"):
        if isinstance(obj, bytes):
            obj = str(obj, encoding, "replace")
        return obj

    @staticmethod
    def _convert_number(org_number):
        """
        Convert org_number into an integer
        org_number: integer or representation of a number: string or str
        Try force converting to int first, on error try converting from Roman numerals
        returns integer or 0
        """

        try:
            # try forcing to int
            if org_number:
                number = int(org_number)
            else:
                number = 0

        except Exception:
            # on error try converting from Roman numerals
            roman_to_int_map = (
                ("M", 1000),
                ("CM", 900),
                ("D", 500),
                ("CD", 400),
                ("C", 100),
                ("XC", 90),
                ("L", 50),
                ("XL", 40),
                ("X", 10),
                ("IX", 9),
                ("V", 5),
                ("IV", 4),
                ("I", 1),
            )

            roman_numeral = str(org_number).upper()
            number = 0
            index = 0

            for numeral, integer in roman_to_int_map:
                while roman_numeral[index : index + len(numeral)] == numeral:
                    number += integer
                    index += len(numeral)

        return number

    def parse(self, name, cache_result=True, skip_scene_detection=False):
        name = self._unicodify(name)

        if self.naming_pattern:
            cache_result = False

        cached = name_parser_cache[name]
        if cached:
            return cached

        # break it into parts if there are any (dirname, file name, extension)
        dir_name, filename = os.path.split(name)

        if self.filename:
            base_filename = remove_extension(filename)
        else:
            base_filename = filename

        # set up a result to use
        final_result = ParseResult(name)

        # try parsing the file name
        filename_result = self._parse_string(base_filename, skip_scene_detection)

        # use only the direct parent dir
        dir_name = os.path.basename(dir_name)

        # parse the dirname for extra info if needed
        dir_name_result = self._parse_string(dir_name, skip_scene_detection)

        # build the ParseResult object
        final_result.air_date = self._combine_results(filename_result, dir_name_result, "air_date")

        # anime absolute numbers
        final_result.ab_episode_numbers = self._combine_results(filename_result, dir_name_result, "ab_episode_numbers")

        # season and episode numbers
        final_result.season_number = self._combine_results(filename_result, dir_name_result, "season_number")
        final_result.episode_numbers = self._combine_results(filename_result, dir_name_result, "episode_numbers")
        final_result.scene_season = self._combine_results(filename_result, dir_name_result, "scene_season")

        # if the dirname has a release group/show name I believe it over the filename
        final_result.series_name = self._combine_results(dir_name_result, filename_result, "series_name")
        final_result.extra_info = self._combine_results(dir_name_result, filename_result, "extra_info")
        final_result.release_group = self._combine_results(dir_name_result, filename_result, "release_group")
        final_result.version = self._combine_results(dir_name_result, filename_result, "version")

        final_result.which_regex = []
        if final_result == filename_result:
            final_result.which_regex = filename_result.which_regex
            final_result.score = filename_result.score
        elif final_result == dir_name_result:
            final_result.which_regex = dir_name_result.which_regex
            final_result.score = dir_name_result.score
        else:
            final_result.score = 0
            if filename_result:
                final_result.which_regex += filename_result.which_regex
                final_result.score += filename_result.score
            if dir_name_result:
                final_result.which_regex += dir_name_result.which_regex
                final_result.score += dir_name_result.score

        final_result.show = self._combine_results(filename_result, dir_name_result, "show")
        final_result.quality = self._combine_results(filename_result, dir_name_result, "quality")

        if not final_result.show:
            raise InvalidShowException(f"Unable to match {name} to a show in your database. Parser result: {final_result}")

        # if there's no useful info in it then raise an exception
        if (
            final_result.season_number is None
            and not final_result.episode_numbers
            and final_result.air_date is None
            and not final_result.ab_episode_numbers
            and not final_result.series_name
        ):
            raise InvalidNameException(f"Unable to parse {name} to a valid episode of {final_result.show.name}. Parser result: {final_result}")

        if cache_result:
            name_parser_cache[name] = final_result

        logger.debug(f"Parsed {name} into {final_result}")
        return final_result


class ParseResult(object):
    def __init__(
        self,
        original_name,
        series_name=None,
        season_number=None,
        episode_numbers=None,
        extra_info=None,
        release_group=None,
        air_date=None,
        ab_episode_numbers=None,
        show=None,
        score=0,
        quality=None,
        version=None,
    ):

        self.original_name = original_name

        self.series_name = series_name
        self.season_number = season_number
        if not episode_numbers:
            self.episode_numbers = []
        else:
            self.episode_numbers = episode_numbers

        if not ab_episode_numbers:
            self.ab_episode_numbers = []
        else:
            self.ab_episode_numbers = ab_episode_numbers

        if not quality:
            self.quality = common.Quality.UNKNOWN
        else:
            self.quality = quality

        self.extra_info = extra_info
        self.release_group = release_group

        self.air_date = air_date

        self.which_regex = []
        self.show: "TVShow" = show
        self.score = score

        self.version = version

        self.scene_season = None

    def __eq__(self, other):
        return other and all(
            [
                self.series_name == other.series_name,
                self.season_number == other.season_number,
                self.episode_numbers == other.episode_numbers,
                self.extra_info == other.extra_info,
                self.release_group == other.release_group,
                self.air_date == other.air_date,
                self.ab_episode_numbers == other.ab_episode_numbers,
                self.show == other.show,
                self.score == other.score,
                self.quality == other.quality,
                self.version == other.version,
            ]
        )

    def __str__(self):
        if self.series_name is not None:
            to_return = f"{self.series_name} - "
        else:
            to_return = ""
        if self.season_number is not None:
            to_return += f"S{self.season_number:02}"
        if self.episode_numbers:
            for e in self.episode_numbers:
                if e is not None:
                    to_return += f"E{e:02}"

        if self.is_air_by_date:
            to_return += f" {self.air_date}"
        if self.ab_episode_numbers:
            to_return += f" [ABS: {self.ab_episode_numbers}]"
        if self.version and self.is_anime is True:
            to_return += f" [ANIME VER: {self.version}]"

        if self.release_group:
            to_return += f" [GROUP: {self.release_group}]"

        to_return += f" [ABD: {self.is_air_by_date}] [ANIME: {self.is_anime}] [whichReg: {self.which_regex}] Score: {self.score}"

        return to_return

    @property
    def is_air_by_date(self):
        return bool(self.air_date)

    @property
    def is_anime(self):
        return bool(self.ab_episode_numbers)


class NameParserCache(object):
    def __init__(self):
        self.lock = Lock()
        self.data = OrderedDict()
        self.max_size = 200

    def __getitem__(self, name):
        with self.lock:
            value = self.data.get(name)
            if value:
                logger.debug(f"Using cached parse result for: {name}")
            return value

    def __setitem__(self, key, value):
        with self.lock:
            self.data.update({key: value})
            while len(self.data) > self.max_size:
                self.data.pop(list(self.data)[0], None)


name_parser_cache = NameParserCache()


class InvalidNameException(Exception):
    """The given release name is not valid"""


class InvalidShowException(Exception):
    """The given show name is not valid"""



# -*- coding: utf-8 -*-
__title__ = 'subliminal'
__version__ = '2.1.0'
__short_version__ = '.'.join(__version__.split('.')[:2])
__author__ = 'Antoine Bertin'
__license__ = 'MIT'
__copyright__ = 'Copyright 2016, Antoine Bertin'

import logging

from .core import (AsyncProviderPool, ProviderPool, check_video, download_best_subtitles, download_subtitles,
                   list_subtitles, refine, save_subtitles, scan_video, scan_videos)
from .cache import region
from .exceptions import Error, ProviderError
from .extensions import provider_manager, refiner_manager
from .providers import Provider
from .score import compute_score, get_scores
from .subtitle import SUBTITLE_EXTENSIONS, Subtitle
from .video import VIDEO_EXTENSIONS, Episode, Movie, Video

logging.getLogger(__name__).addHandler(logging.NullHandler())



# TODO: Implement skipped tests

import unittest

import pytest

from sickchill import settings
from sickchill.oldbeard import common


class QualityStringTests(unittest.TestCase):
    """
    Test Case for strings in common.Quality
    """

    settings.QUALITY_ALLOW_HEVC = True
    test_cases = {
        "sd_tv": [
            "Test.Show.S01E02.PDTV.XViD-GROUP",
            "Test.Show.S01E02.PDTV.x264-GROUP",
            "Test.Show.S01E02.HDTV.XViD-GROUP",
            "Test.Show.S01E02.HDTV.x264-GROUP",
            "Test.Show.S01E02.DSR.XViD-GROUP",
            "Test.Show.S01E02.DSR.x264-GROUP",
            "Test.Show.S01E02.TVRip.XViD-GROUP",
            "Test.Show.S01E02.TVRip.x264-GROUP",
            "Test.Show.S01E02.WEBRip.XViD-GROUP",
            "Test.Show.S01E02.WEBRip.x264-GROUP",
            "Test.Show.S01E02.WEB-DL.x264-GROUP",
            "Test.Show.S01E02.WEB-DL.AAC2.0.H.264-GROUP",
            "Test.Show.S01E02 WEB-DL H 264-GROUP",
            "Test.Show.S01E02_WEB-DL_H_264-GROUP",
            "Test.Show.S01E02.WEB-DL.AAC2.0.H264-GROUP",
        ],
        "sd_dvd": [
            "Test.Show.S01E02.480P.DVDrip.HEVC.X265",
            "Test.Show.S01E02.DVDRiP.XViD-GROUP",
            "Test.Show.S01E02.DVDRiP.DiVX-GROUP",
            "Test.Show.S01E02.DVDRiP.x264-GROUP",
            "Test.Show.S01E02.DVDRip.WS.XViD-GROUP",
            "Test.Show.S01E02.DVDRip.WS.DiVX-GROUP",
            "Test.Show.S01E02.DVDRip.WS.x264-GROUP",
            "Test.Show.S01E02.BDRIP.XViD-GROUP",
            "Test.Show.S01E02.BDRIP.DiVX-GROUP",
            "Test.Show.S01E02.BDRIP.x264-GROUP",
            "Test.Show.S01E02.BDRIP.WS.XViD-GROUP",
            "Test.Show.S01E02.BDRIP.WS.DiVX-GROUP",
            "Test.Show.S01E02.BDRIP.WS.x264-GROUP",
        ],
        "hd_tv": ["Test.Show.S01E02.720p.HDTV.x264-GROUP", "Test.Show.S01E02.HR.WS.PDTV.x264-GROUP"],
        "raw_hd_tv": [
            "Test.Show.S01E02.720p.HDTV.DD5.1.MPEG2-GROUP",
            "Test.Show.S01E02.1080i.HDTV.DD2.0.MPEG2-GROUP",
            "Test.Show.S01E02.1080i.HDTV.H.264.DD2.0-GROUP",
            "Test Show - S01E02 - 1080i HDTV MPA1.0 H.264 - GROUP",
            "Test.Show.S01E02.1080i.HDTV.DD.5.1.h264-GROUP",
        ],
        "full_hd_tv": ["Test.Show.S01E02.1080p.HDTV.x264-GROUP"],
        "hd_web_dl": [
            "Test.Show.S01E02.720p.WEB-DL-GROUP",
            "Test.Show.S01E02.720p.WEBRip-GROUP",
            "Test.Show.S01E02.WEBRip.720p.H.264.AAC.2.0-GROUP",
            "Test.Show.S01E02.720p.WEB-DL.AAC2.0.H.264-GROUP",
            "Test Show S01E02 720p WEB-DL AAC2 0 H 264-GROUP",
            "Test_Show.S01E02_720p_WEB-DL_AAC2.0_H264-GROUP",
            "Test.Show.S01E02.720p.WEB-DL.AAC2.0.H264-GROUP",
            "Test.Show.S01E02.720p.iTunes.Rip.H264.AAC-GROUP",
            "Test.Show.S01E02.Episode.Name.Itunes.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.ItunesHD.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.ItunesUHD.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.720p.Itunes.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.720p.ItunesHD.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.720p.ItunesUHD.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.AMZN.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.Amazon.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.AmazonHD.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.AmazonUHD.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.720p.AMZN.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.720p.Amazon.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.720p.AmazonHD.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.720p.AmazonUHD.WEB-DL.x264",
            # German dubs
            "Test.Show.S01E02.Episode.Name.German.DD51.Synced.DL.iTunesHD.AVC",
            "Test.Show.S01E02.Episode.Name.German.DD51.Synced.DL.AmazonHD.x264",
            "Test.Show.S01E02.Episode.Name.German.Dubbed.DL.iTunesHD.x264",
            "Test.Show.S01E02.Episode.Name.German.DD51.DL.NetflixHD.x264",
            "Test.Show.S01E02.Episode.Name.German.DD51.DL.NetflixUHD.x264",
        ],
        "full_hd_web_dl": [
            "Test.Show.S01E02.1080p.WEB-DL-GROUP",
            "Test.Show.S01E02.1080p.WEBRip-GROUP",
            "Test.Show.S01E02.WEBRip.1080p.H.264.AAC.2.0-GROUP",
            "Test.Show.S01E02.WEBRip.1080p.H264.AAC.2.0-GROUP",
            "Test.Show.S01E02.1080p.iTunes.H.264.AAC-GROUP",
            "Test Show S01E02 1080p iTunes H 264 AAC-GROUP",
            "Test_Show_S01E02_1080p_iTunes_H_264_AAC-GROUP",
            "Test.Show.S01E02.Episode.Name.1080p.Itunes.WEB-DL.x264" "Test.Show.S01E02.Episode.Name.1080p.ItunesHD.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.1080p.ItunesUHD.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.1080p.AMZN.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.1080p.Amazon.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.1080p.AmazonUHD.WEB-DL.x264",
            "Test.Show.S01E02.Episode.Name.1080p.AmazonHD.WEB-DL.x264",
            # German dubs
            "Test.Show.S01E02.Episode.Name.German.DD51.Synced.DL.1080p.iTunesHD.AVC",
            "Test.Show.S01E02.Episode.Name.German.DD51.Synced.DL.1080p.AmazonHD.x264",
            "Test.Show.S01E02.Episode.Name.German.Dubbed.DL.1080p.iTunesHD.x264",
            "Test.Show.S01E02.Episode.Name.German.DD51.DL.1080p.NetflixHD.x264",
            "Test.Show.S01E02.Episode.Name.German.DD51.DL.1080p.NetflixUHD.x264",
        ],
        "hd_bluray": ["Test.Show.S01E02.720p.BluRay.x264-GROUP", "Test.Show.S01E02.720p.HDDVD.x264-GROUP"],
        "full_hd_bluray": ["Test.Show.S01E02.1080p.BluRay.x264-GROUP", "Test.Show.S01E02.1080p.HDDVD.x264-GROUP"],
        "unknown": [
            "Test.Show.S01E02-SiCKCHILL",
            "Test.Show.S01E01-20.1080i.[Mux.-.1080i.-.H264.-.Ac3.].HDTVMux.GROUP",
        ],
    }

    def test_sd_tv(self):
        """
        Test SDTV against nameQuality
        """
        cur_test = "sd_tv"
        cur_qual = common.Quality.SDTV

        for name, tests in self.test_cases.items():
            for test in tests:
                if name == cur_test:
                    assert cur_qual == common.Quality.nameQuality(test)
                else:
                    assert cur_qual != common.Quality.nameQuality(test)

    def test_sd_dvd(self):
        """
        Test SDDVD against nameQuality
        """
        cur_test = "sd_dvd"
        cur_qual = common.Quality.SDDVD

        for name, tests in self.test_cases.items():
            for test in tests:
                if name == cur_test:
                    assert cur_qual == common.Quality.nameQuality(test)
                else:
                    assert cur_qual != common.Quality.nameQuality(test)

    def test_hd_tv(self):
        """
        Test HDTV against nameQuality
        """
        cur_test = "hd_tv"
        cur_qual = common.Quality.HDTV

        for name, tests in self.test_cases.items():
            for test in tests:
                if name == cur_test:
                    assert cur_qual == common.Quality.nameQuality(test), test
                else:
                    assert cur_qual != common.Quality.nameQuality(test), test

    def test_raw_hd_tv(self):
        """
        Test RAWHDTV against nameQuality
        """
        cur_test = "raw_hd_tv"
        cur_qual = common.Quality.RAWHDTV

        for name, tests in self.test_cases.items():
            for test in tests:
                if name == cur_test:
                    assert cur_qual == common.Quality.nameQuality(test), test
                else:
                    assert cur_qual != common.Quality.nameQuality(test), test

    def test_full_hd_tv(self):
        """
        Test FULLHDTV against nameQuality
        """
        cur_test = "full_hd_tv"
        cur_qual = common.Quality.FULLHDTV

        for name, tests in self.test_cases.items():
            for test in tests:
                if name == cur_test:
                    assert cur_qual == common.Quality.nameQuality(test)
                else:
                    assert cur_qual != common.Quality.nameQuality(test)

    def test_hd_web_dl(self):
        """
        Test HDWEBDL against nameQuality
        """
        cur_test = "hd_web_dl"
        cur_qual = common.Quality.HDWEBDL

        for name, tests in self.test_cases.items():
            for test in tests:
                if name == cur_test:
                    assert cur_qual == common.Quality.nameQuality(test)
                else:
                    assert cur_qual != common.Quality.nameQuality(test)

    def test_full_hd_web_dl(self):
        """
        Test FULLHDWEBDL against nameQuality
        """
        cur_test = "full_hd_web_dl"
        cur_qual = common.Quality.FULLHDWEBDL

        for name, tests in self.test_cases.items():
            for test in tests:
                if name == cur_test:
                    assert cur_qual == common.Quality.nameQuality(test)
                else:
                    assert cur_qual != common.Quality.nameQuality(test)

    def test_hd_bluray(self):
        """
        Test HDBLURAY against nameQuality
        """
        cur_test = "hd_bluray"
        cur_qual = common.Quality.HDBLURAY

        for name, tests in self.test_cases.items():
            for test in tests:
                if name == cur_test:
                    assert cur_qual == common.Quality.nameQuality(test)
                else:
                    assert cur_qual != common.Quality.nameQuality(test)

    def test_full_hd_bluray(self):
        """
        Test FULLHDBLURAY against nameQuality
        """
        cur_test = "full_hd_bluray"
        cur_qual = common.Quality.FULLHDBLURAY

        for name, tests in self.test_cases.items():
            for test in tests:
                if name == cur_test:
                    assert cur_qual == common.Quality.nameQuality(test)
                else:
                    assert cur_qual != common.Quality.nameQuality(test)

    def test_unknown(self):
        """
        Test UNKNOWN against nameQuality
        """
        cur_test = "unknown"
        cur_qual = common.Quality.UNKNOWN

        for name, tests in self.test_cases.items():
            for test in tests:
                if name == cur_test:
                    assert cur_qual == common.Quality.nameQuality(test)
                else:
                    assert cur_qual != common.Quality.nameQuality(test)

    def test_anime(self):
        """
        Test Anime against nameQuality
        """
        test_cases = {
            "sd_dvd": [
                "[DeadFish].Shingeki.no.Kyojin.-.01.-.OVA.[DVD][480p][AAC].mp4",
            ],
            "full_hd_bluray": [
                "[Hatsuyuki-Kaitou]_Shingeki_no_Kyojin_-_Special_02_[BD_1080p][10bit][FLAC][4D908801].mkv",
            ],
        }
        test_quality = {
            "sd_tv": common.Quality.SDTV,
            "sd_dvd": common.Quality.SDDVD,
            "raw_hd_tv": common.Quality.RAWHDTV,
            "hd_tv": common.Quality.HDTV,
            "hd_web_dl": common.Quality.HDWEBDL,
            "hd_bluray": common.Quality.HDBLURAY,
            "full_hd_tv": common.Quality.FULLHDTV,
            "full_hd_web_dl": common.Quality.FULLHDWEBDL,
            "full_hd_bluray": common.Quality.FULLHDBLURAY,
            "unknown": common.Quality.UNKNOWN,
        }
        for cur_test, expected_qual in test_quality.items():
            for qual, tests in test_cases.items():
                for name in tests:
                    name_quality = common.Quality.nameQuality(name, anime=True)
                    name_info = (qual, name, expected_qual, common.Quality.nameQuality(name, anime=True))
                    if qual == cur_test:
                        assert expected_qual == name_quality, name_info
                    else:
                        assert expected_qual != name_quality, name_info


class QualityTests(unittest.TestCase):
    """
    Test Case for common.Quality
    """

    # TODO: repack / proper ? air-by-date ? season rip? multi-ep?
    # reverse parsing does not work
    def test_reverse_parsing(self):
        """
        Test reverse parsing for all qualities
        """
        tests = [
            (common.Quality.SDTV, "Test Show - S01E02 - SDTV - GROUP"),
            (common.Quality.SDDVD, "Test Show - S01E02 - SD DVD - GROUP"),
            (common.Quality.HDTV, "Test Show - S01E02 - HDTV - GROUP"),
            (common.Quality.RAWHDTV, "Test Show - S01E02 - RawHD - GROUP"),
            (common.Quality.FULLHDTV, "Test Show - S01E02 - 1080p HDTV - GROUP"),
            (common.Quality.HDWEBDL, "Test Show - S01E02 - 720p WEB-DL - GROUP"),
            (common.Quality.FULLHDWEBDL, "Test Show - S01E02 - 1080p WEB-DL - GROUP"),
            (common.Quality.HDBLURAY, "Test Show - S01E02 - 720p BluRay - GROUP"),
            (common.Quality.FULLHDBLURAY, "Test Show - S01E02 - 1080p BluRay - GROUP"),
            (common.Quality.UNKNOWN, "Test Show - S01E02 - Unknown - SiCKCHILL"),
        ]
        for test in tests:
            quality, test = test
            assert quality == common.Quality.nameQuality(test), (quality, common.Quality.nameQuality(test), test)

    @unittest.skip("Not yet implemented")
    def test_get_status_strings(self):
        """
        Test _getStatusStrings
        """
        pass

    @unittest.skip("Not yet implemented")
    def test_combine_qualities(self):
        """
        Test combineQualities
        """
        pass

    @unittest.skip("Not yet implemented")
    def test_split_quality(self):
        """
        Test splitQuality
        """
        pass

    @unittest.skip("Not yet implemented")
    def test_name_quality(self):
        """
        Test nameQuality
        """
        pass

    @unittest.skip("Not yet implemented")
    def test_scene_quality(self):
        """
        Test scene_quality
        """
        pass

    @unittest.skip("Not yet implemented")
    def test_quality_from_file_meta(self):
        """
        Test qualityFromFileMeta
        """
        pass

    @unittest.skip("Not yet implemented")
    def test_composite_status(self):
        """
        Test compositeStatus
        """
        pass

    @unittest.skip("Not yet implemented")
    def test_quality_downloaded(self):
        """
        Test qualityDownloaded
        """
        pass

    @unittest.skip("Not yet implemented")
    def test_split_composite_status(self):
        """
        Test splitCompositeStatus
        """
        pass

    @unittest.skip("Not yet implemented")
    def test_scene_quality_from_name(self):
        """
        Test sceneQualityFromName
        """
        pass

    @unittest.skip("Not yet implemented")
    def test_status_from_name(self):
        """
        Test statusFromName
        """
        pass


class StatusStringsTests(unittest.TestCase):
    """
    Test Case for common.StatusStrings
    """

    # TODO: Split tests into separate tests and add additional tests

    def test_all(self):
        """
        Run all status strings tests
        """
        status_strings = common.statusStrings

        valid = 1, 112, "1", "112"
        unused = 122, 99998989899878676, "99998989899878676", None
        invalid = "Elephant", (4, 1), [1, 233, 4, None]

        for i in valid:
            assert i in status_strings

        for i in unused:
            assert i not in status_strings
            with pytest.raises(KeyError):
                status_strings[i]

        for i in status_strings:
            assert status_strings[i] == status_strings[str(i)]
            assert (i in status_strings) == (str(i) in status_strings)

        for i in status_strings.qualities:
            assert status_strings[i] == status_strings[str(i)]
            assert (i in status_strings) == (str(i) in status_strings)

        for i in invalid:
            with self.assertRaises(TypeError):
                status_strings[i] = 1

        for i in unused:
            if i is None:
                with self.assertRaises(TypeError):
                    status_strings[str(i)] = 1  # 'None' is not None
                status_strings[i] = 1  # ...but None can still be used as a key
            else:
                status_strings[str(i)] = 1
            assert status_strings[i] == 1


class OverviewTests(unittest.TestCase):
    """
    Test common.Overview
    """

    def test_overview_strings(self):
        """
        Test common.Overview.overviewStrings
        """
        overview = common.Overview()

        assert overview.overviewStrings[overview.SKIPPED] == "skipped"
        assert overview.overviewStrings[overview.WANTED] == "wanted"
        assert overview.overviewStrings[overview.QUAL] == "qual"
        assert overview.overviewStrings[overview.GOOD] == "good"
        assert overview.overviewStrings[overview.UNAIRED] == "unaired"
        assert overview.overviewStrings[overview.SNATCHED] == "snatched"


if __name__ == "__main__":
    print("=======================")
    print("STARTING - COMMON TESTS")
    print("=======================")

    SUITE = unittest.TestLoader().loadTestsFromTestCase(QualityStringTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)

    SUITE = unittest.TestLoader().loadTestsFromTestCase(QualityTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)

    SUITE = unittest.TestLoader().loadTestsFromTestCase(StatusStringsTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)

    SUITE = unittest.TestLoader().loadTestsFromTestCase(OverviewTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



# Copyright: 2013 Paul Traylor
# These sources are released under the terms of the MIT license: see LICENSE

"""
Python2.5 and Python3.3 compatibility shim

Heavily inspirted by the "six" library.
https://pypi.python.org/pypi/six
"""

import sys

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

if PY3:
	def b(s):
		if isinstance(s, bytes):
			return s
		return s.encode('utf8', 'replace')

	def u(s):
		if isinstance(s, bytes):
			return s.decode('utf8', 'replace')
		return s

	from io import BytesIO as StringIO
	from configparser import RawConfigParser
else:
	def b(s):
		if isinstance(s, unicode):
			return s.encode('utf8', 'replace')
		return s

	def u(s):
		if isinstance(s, unicode):
			return s
		if isinstance(s, int):
			s = str(s)
		return unicode(s, "utf8", "replace")

	from StringIO import StringIO
	from ConfigParser import RawConfigParser

b.__doc__ = "Ensure we have a byte string"
u.__doc__ = "Ensure we have a unicode string"



import json
import os

from tornado.web import addslash

import sickchill.start
from sickchill import settings
from sickchill.helper import try_int
from sickchill.oldbeard import config, filters, ui
from sickchill.oldbeard.providers import newznab, rsstorrent
from sickchill.providers.GenericProvider import GenericProvider
from sickchill.views.common import PageTemplate
from sickchill.views.routes import Route

from . import Config


@Route("/config/providers(/?.*)", name="config:providers")
class ConfigProviders(Config):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @addslash
    def index(self, *args_, **kwargs_):
        t = PageTemplate(rh=self, filename="config_providers.mako")

        return t.render(
            submenu=self.ConfigMenu(), title=_("Config - Providers"), header=_("Search Providers"), topmenu="config", controller="config", action="providers"
        )

    @staticmethod
    def canAddNewznabProvider(name):

        if not name:
            return json.dumps({"error": "No Provider Name specified"})

        providerDict = {x.get_id(): x for x in settings.newznabProviderList}

        cur_id = GenericProvider.make_id(name)

        if cur_id in providerDict:
            return json.dumps({"error": "Provider Name already exists as " + name})
        else:
            return json.dumps({"success": cur_id})

    @staticmethod
    def getNewznabCategories(name, url, key):
        """
        Retrieves a list of possible categories with category id's
        Using the default url/api?cat
        http://yournewznaburl.com/api?t=caps&apikey=yourapikey
        """
        error = ""

        if not name:
            error += "\n" + _("No Provider Name specified")
        if not url:
            error += "\n" + _("No Provider Url specified")
        if not key:
            error += "\n" + _("No Provider Api key specified")

        if error:
            return json.dumps({"success": False, "error": error})

        # Get newznabprovider obj with provided name
        tempProvider = newznab.NewznabProvider(name, url, key)

        success, tv_categories, error = tempProvider.get_newznab_categories()

        return json.dumps({"success": success, "tv_categories": tv_categories, "error": error})

    @staticmethod
    def deleteNewznabProvider(nnid):

        providerDict = {x.get_id(): x for x in settings.newznabProviderList}
        if nnid not in providerDict or providerDict[nnid].default:
            return "0"

        # delete it from the list
        settings.newznabProviderList.remove(providerDict[nnid])

        if nnid in settings.PROVIDER_ORDER:
            settings.PROVIDER_ORDER.remove(nnid)

        return "1"

    @staticmethod
    def canAddTorrentRssProvider(name, url, cookies, titleTAG):

        if not name:
            return json.dumps({"error": "Invalid name specified"})

        url = config.clean_url(url)
        tempProvider = rsstorrent.TorrentRssProvider(name, url, cookies, titleTAG)

        if tempProvider.get_id() in (x.get_id() for x in settings.torrentRssProviderList):
            return json.dumps({"error": "Exists as " + tempProvider.name})
        else:
            (succ, errMsg) = tempProvider.validateRSS()
            if succ:
                return json.dumps({"success": tempProvider.get_id()})
            else:
                return json.dumps({"error": errMsg})

    @staticmethod
    def deleteTorrentRssProvider(provider_id):

        providerDict = {x.get_id(): x for x in settings.torrentRssProviderList}

        if provider_id not in providerDict:
            return "0"

        # delete it from the list
        settings.torrentRssProviderList.remove(providerDict[provider_id])

        if provider_id in settings.PROVIDER_ORDER:
            settings.PROVIDER_ORDER.remove(provider_id)

        return "1"

    def saveProviders(self, newznab_string="", torrentrss_string="", provider_order=None, **kwargs):
        newznabProviderDict = {x.get_id(): x for x in settings.newznabProviderList}

        finished_names = []

        # add all the newznab info we got into our list
        # if not newznab_string:
        #     logger.debug('No newznab_string passed to saveProviders')

        for curNewznabProviderStr in newznab_string.split("!!!"):
            if not curNewznabProviderStr:
                continue

            cur_name, cur_url, cur_key, cur_cat = curNewznabProviderStr.split("|")
            cur_url = config.clean_url(cur_url)
            cur_id = GenericProvider.make_id(cur_name)

            # if it does not already exist then add it
            if cur_id not in newznabProviderDict:
                new_provider = newznab.NewznabProvider(cur_name, cur_url, key=cur_key, catIDs=cur_cat)
                settings.newznabProviderList.append(new_provider)
                newznabProviderDict[cur_id] = new_provider

            # set all params
            newznabProviderDict[cur_id].name = cur_name
            newznabProviderDict[cur_id].url = cur_url
            newznabProviderDict[cur_id].key = cur_key
            newznabProviderDict[cur_id].catIDs = cur_cat
            # a 0 in the key spot indicates that no key is needed
            newznabProviderDict[cur_id].needs_auth = cur_key and cur_key != "0"
            newznabProviderDict[cur_id].search_mode = str(kwargs.get(cur_id + "_search_mode", "eponly")).strip()
            newznabProviderDict[cur_id].search_fallback = config.checkbox_to_value(kwargs.get(cur_id + "search_fallback", 0), value_on=1, value_off=0)
            newznabProviderDict[cur_id].enable_daily = config.checkbox_to_value(kwargs.get(cur_id + "enable_daily", 0), value_on=1, value_off=0)
            newznabProviderDict[cur_id].enable_backlog = config.checkbox_to_value(kwargs.get(cur_id + "enable_backlog", 0), value_on=1, value_off=0)

            # mark it finished
            finished_names.append(cur_id)

        # delete anything that is in the list that was not processed just now
        if newznab_string:
            for curProvider in settings.newznabProviderList:
                if curProvider.get_id() not in finished_names:
                    settings.newznabProviderList.remove(curProvider)
                    del newznabProviderDict[curProvider.get_id()]

        # if not torrentrss_string:
        #     logger.debug('No torrentrss_string passed to saveProviders')

        torrentRssProviderDict = {x.get_id(): x for x in settings.torrentRssProviderList}

        finished_names = []

        if torrentrss_string:
            for curTorrentRssProviderStr in torrentrss_string.split("!!!"):

                if not curTorrentRssProviderStr:
                    continue

                cur_name, cur_url, cur_cookies, cur_title_tag = curTorrentRssProviderStr.split("|")
                cur_url = config.clean_url(cur_url)
                cur_id = GenericProvider.make_id(cur_name)

                # if it does not already exist then create it
                if cur_id not in torrentRssProviderDict:
                    new_provider = rsstorrent.TorrentRssProvider(cur_name, cur_url, cur_cookies, cur_title_tag)
                    settings.torrentRssProviderList.append(new_provider)
                    torrentRssProviderDict[cur_id] = new_provider

                # update values
                torrentRssProviderDict[cur_id].name = cur_name
                torrentRssProviderDict[cur_id].url = cur_url
                torrentRssProviderDict[cur_id].cookies = cur_cookies
                torrentRssProviderDict[cur_id].cur_title_tag = cur_title_tag

                # mark it finished
                finished_names.append(cur_id)

        # delete anything that is in the list that was not processed just now
        if torrentrss_string:
            for curProvider in settings.torrentRssProviderList:
                if curProvider.get_id() not in finished_names:
                    settings.torrentRssProviderList.remove(curProvider)
                    del torrentRssProviderDict[curProvider.get_id()]

        # do the enable/disable
        enabled_provider_list = []
        disabled_provider_list = []
        for cur_id, cur_enabled in (cur_provider_str.split(":") for cur_provider_str in provider_order.split()):
            cur_enabled = bool(try_int(cur_enabled))

            cur_provider_obj = [x for x in sickchill.oldbeard.providers.sortedProviderList() if x.get_id() == cur_id and hasattr(x, "enabled")]

            if cur_provider_obj:
                cur_provider_obj[0].enabled = cur_enabled

            if cur_enabled:
                enabled_provider_list.append(cur_id)
            else:
                disabled_provider_list.append(cur_id)

            if cur_id in newznabProviderDict:
                newznabProviderDict[cur_id].enabled = cur_enabled
            elif cur_id in torrentRssProviderDict:
                torrentRssProviderDict[cur_id].enabled = cur_enabled

        # dynamically load provider settings
        for curProvider in sickchill.oldbeard.providers.sortedProviderList():
            if hasattr(curProvider, "custom_url"):
                curProvider.custom_url = str(kwargs.get(curProvider.get_id("_custom_url"), "")).strip()

            if hasattr(curProvider, "minseed"):
                curProvider.minseed = int(str(kwargs.get(curProvider.get_id("_minseed"), 0)).strip())

            if hasattr(curProvider, "minleech"):
                curProvider.minleech = int(str(kwargs.get(curProvider.get_id("_minleech"), 0)).strip())

            if hasattr(curProvider, "ratio"):
                if curProvider.get_id("_ratio") in kwargs:
                    ratio = str(kwargs.get(curProvider.get_id("_ratio"))).strip()
                    if ratio in ("None", None, ""):
                        curProvider.ratio = None
                    else:
                        curProvider.ratio = max(float(ratio), -1)
                else:
                    curProvider.ratio = None

            if hasattr(curProvider, "digest"):
                curProvider.digest = str(kwargs.get(curProvider.get_id("_digest"), "")).strip() or None

            if hasattr(curProvider, "hash"):
                curProvider.hash = str(kwargs.get(curProvider.get_id("_hash"), "")).strip() or None

            if hasattr(curProvider, "api_key"):
                curProvider.api_key = str(kwargs.get(curProvider.get_id("_api_key"), "")).strip() or None

            if hasattr(curProvider, "username"):
                curProvider.username = str(kwargs.get(curProvider.get_id("_username"), "")).strip() or None

            if hasattr(curProvider, "password"):
                curProvider.password = filters.unhide(curProvider.password, str(kwargs.get(curProvider.get_id("_password"), "")).strip())

            if hasattr(curProvider, "passkey"):
                curProvider.passkey = filters.unhide(curProvider.passkey, str(kwargs.get(curProvider.get_id("_passkey"), "")).strip())

            if hasattr(curProvider, "pin"):
                curProvider.pin = filters.unhide(curProvider.pin, str(kwargs.get(curProvider.get_id("_pin"), "")).strip())

            if hasattr(curProvider, "confirmed"):
                curProvider.confirmed = config.checkbox_to_value(kwargs.get(curProvider.get_id("_confirmed")))

            if hasattr(curProvider, "ranked"):
                curProvider.ranked = config.checkbox_to_value(kwargs.get(curProvider.get_id("_ranked")))

            if hasattr(curProvider, "engrelease"):
                curProvider.engrelease = config.checkbox_to_value(kwargs.get(curProvider.get_id("_engrelease")))

            if hasattr(curProvider, "onlyspasearch"):
                curProvider.onlyspasearch = config.checkbox_to_value(kwargs.get(curProvider.get_id("_onlyspasearch")))

            if hasattr(curProvider, "sorting"):
                curProvider.sorting = str(kwargs.get(curProvider.get_id("_sorting"), "seeders")).strip()

            if hasattr(curProvider, "freeleech"):
                curProvider.freeleech = config.checkbox_to_value(kwargs.get(curProvider.get_id("_freeleech")))

            if hasattr(curProvider, "search_mode"):
                curProvider.search_mode = str(kwargs.get(curProvider.get_id("_search_mode"), "eponly")).strip()

            if hasattr(curProvider, "search_fallback"):
                curProvider.search_fallback = config.checkbox_to_value(kwargs.get(curProvider.get_id("_search_fallback")))

            if hasattr(curProvider, "enable_daily"):
                curProvider.enable_daily = curProvider.can_daily and config.checkbox_to_value(kwargs.get(curProvider.get_id("_enable_daily")))

            if hasattr(curProvider, "enable_backlog"):
                curProvider.enable_backlog = curProvider.can_backlog and config.checkbox_to_value(kwargs.get(curProvider.get_id("_enable_backlog")))

            if hasattr(curProvider, "cat"):
                curProvider.cat = int(str(kwargs.get(curProvider.get_id("_cat"), 0)).strip())

            if hasattr(curProvider, "subtitle"):
                curProvider.subtitle = config.checkbox_to_value(kwargs.get(curProvider.get_id("_subtitle")))

            if curProvider.enable_cookies:
                curProvider.cookies = str(kwargs.get(curProvider.get_id("_cookies"))).strip()

        settings.NEWZNAB_DATA = "!!!".join([x.configStr() for x in settings.newznabProviderList])
        settings.PROVIDER_ORDER = enabled_provider_list + disabled_provider_list

        sickchill.start.save_config()

        # Add a site_message if no providers are enabled for daily and/or backlog
        sickchill.oldbeard.providers.check_enabled_providers()

        ui.notifications.message(_("Configuration Saved"), os.path.join(settings.CONFIG_FILE))

        return self.redirect("/config/providers/")



# -*- coding: utf-8 -*-
from __future__ import division
from datetime import datetime, timedelta
import logging
import os

from guessit import guessit
from rebulk.loose import ensure_list

from subliminal.utils import matches_title

logger = logging.getLogger(__name__)

#: Video extensions
VIDEO_EXTENSIONS = ('.3g2', '.3gp', '.3gp2', '.3gpp', '.60d', '.ajp', '.asf', '.asx', '.avchd', '.avi', '.bik',
                    '.bix', '.box', '.cam', '.dat', '.divx', '.dmf', '.dv', '.dvr-ms', '.evo', '.flc', '.fli',
                    '.flic', '.flv', '.flx', '.gvi', '.gvp', '.h264', '.m1v', '.m2p', '.m2ts', '.m2v', '.m4e',
                    '.m4v', '.mjp', '.mjpeg', '.mjpg', '.mk3d', '.mkv', '.moov', '.mov', '.movhd', '.movie', '.movx',
                    '.mp4', '.mpe', '.mpeg', '.mpg', '.mpv', '.mpv2', '.mxf', '.nsv', '.nut', '.ogg', '.ogm', '.ogv',
                    '.omf', '.ps', '.qt', '.ram', '.rm', '.rmvb', '.swf', '.ts', '.vfw', '.vid', '.video', '.viv',
                    '.vivo', '.vob', '.vro', '.webm', '.wm', '.wmv', '.wmx', '.wrap', '.wvx', '.wx', '.x264', '.xvid')


class Video(object):
    """Base class for videos.

    Represent a video, existing or not.

    :param str name: name or path of the video.
    :param str source: source of the video (HDTV, Web, Blu-ray, ...).
    :param str release_group: release group of the video.
    :param str streaming_service: streaming_service of the video.
    :param str resolution: resolution of the video stream (480p, 720p, 1080p or 1080i).
    :param str video_codec: codec of the video stream.
    :param str audio_codec: codec of the main audio stream.
    :param str imdb_id: IMDb id of the video.
    :param dict hashes: hashes of the video file by provider names.
    :param int size: size of the video file in bytes.
    :param set subtitle_languages: existing subtitle languages.

    """
    def __init__(self, name, source=None, release_group=None, resolution=None, streaming_service=None,
                 video_codec=None, audio_codec=None, imdb_id=None, hashes=None, size=None, subtitle_languages=None):
        #: Name or path of the video
        self.name = name

        #: Source of the video (HDTV, Web, Blu-ray, ...)
        self.source = source

        #: Release group of the video
        self.release_group = release_group

        #: Streaming service of the video
        self.streaming_service = streaming_service

        #: Resolution of the video stream (480p, 720p, 1080p or 1080i)
        self.resolution = resolution

        #: Codec of the video stream
        self.video_codec = video_codec

        #: Codec of the main audio stream
        self.audio_codec = audio_codec

        #: IMDb id of the video
        self.imdb_id = imdb_id

        #: Hashes of the video file by provider names
        self.hashes = hashes or {}

        #: Size of the video file in bytes
        self.size = size

        #: Existing subtitle languages
        self.subtitle_languages = subtitle_languages or set()

    @property
    def exists(self):
        """Test whether the video exists"""
        return os.path.exists(self.name)

    @property
    def age(self):
        """Age of the video"""
        if self.exists:
            return datetime.utcnow() - datetime.utcfromtimestamp(os.path.getmtime(self.name))

        return timedelta()

    @classmethod
    def fromguess(cls, name, guess):
        """Create an :class:`Episode` or a :class:`Movie` with the given `name` based on the `guess`.

        :param str name: name of the video.
        :param dict guess: guessed data.
        :raise: :class:`ValueError` if the `type` of the `guess` is invalid

        """
        if guess['type'] == 'episode':
            return Episode.fromguess(name, guess)

        if guess['type'] == 'movie':
            return Movie.fromguess(name, guess)

        raise ValueError('The guess must be an episode or a movie guess')

    @classmethod
    def fromname(cls, name):
        """Shortcut for :meth:`fromguess` with a `guess` guessed from the `name`.

        :param str name: name of the video.

        """
        return cls.fromguess(name, guessit(name))

    def __repr__(self):
        return '<%s [%r]>' % (self.__class__.__name__, self.name)

    def __hash__(self):
        return hash(self.name)


class Episode(Video):
    """Episode :class:`Video`.

    :param str series: series of the episode.
    :param int season: season number of the episode.
    :param int or list episodes: episode numbers of the episode.
    :param str title: title of the episode.
    :param int year: year of the series.
    :param country: Country of the series.
    :type country: :class:`~babelfish.country.Country`
    :param bool original_series: whether the series is the first with this name.
    :param int tvdb_id: TVDB id of the episode.
    :param list alternative_series: alternative names of the series
    :param \*\*kwargs: additional parameters for the :class:`Video` constructor.

    """
    def __init__(self, name, series, season, episodes, title=None, year=None, country=None, original_series=True,
                 tvdb_id=None, series_tvdb_id=None, series_imdb_id=None, alternative_series=None, **kwargs):
        super(Episode, self).__init__(name, **kwargs)

        #: Series of the episode
        self.series = series

        #: Season number of the episode
        self.season = season

        #: Episode numbers of the episode
        self.episodes = ensure_list(episodes)

        #: Title of the episode
        self.title = title

        #: Year of series
        self.year = year

        #: The series is the first with this name
        self.original_series = original_series

        #: Country of the series
        self.country = country

        #: TVDB id of the episode
        self.tvdb_id = tvdb_id

        #: TVDB id of the series
        self.series_tvdb_id = series_tvdb_id

        #: IMDb id of the series
        self.series_imdb_id = series_imdb_id

        #: Alternative names of the series
        self.alternative_series = alternative_series or []

    @property
    def episode(self):
        return min(self.episodes) if self.episodes else None

    def matches(self, series):
        return matches_title(series, self.series, self.alternative_series)

    @classmethod
    def fromguess(cls, name, guess):
        if guess['type'] != 'episode':
            raise ValueError('The guess must be an episode guess')

        if 'title' not in guess or 'episode' not in guess:
            raise ValueError('Insufficient data to process the guess')

        return cls(name, guess['title'], guess.get('season', 1), guess.get('episode'), title=guess.get('episode_title'),
                   year=guess.get('year'), country=guess.get('country'),
                   original_series='year' not in guess and 'country' not in guess,
                   source=guess.get('source'), alternative_series=ensure_list(guess.get('alternative_title')),
                   release_group=guess.get('release_group'), streaming_service=guess.get('streaming_service'),
                   resolution=guess.get('screen_size'),
                   video_codec=guess.get('video_codec'), audio_codec=guess.get('audio_codec'))

    @classmethod
    def fromname(cls, name):
        return cls.fromguess(name, guessit(name, {'type': 'episode'}))

    def __repr__(self):
        return '<{cn} [{series}{open}{country}{sep}{year}{close} s{season:02d}e{episodes}]>'.format(
            cn=self.__class__.__name__, series=self.series, year=self.year or '', country=self.country or '',
            season=self.season, episodes='-'.join(map(lambda v: '{:02d}'.format(v), self.episodes)),
            open=' (' if not self.original_series else '',
            sep=') (' if self.year and self.country else '',
            close=')' if not self.original_series else ''
        )


class Movie(Video):
    """Movie :class:`Video`.

    :param str title: title of the movie.
    :param int year: year of the movie.
    :param country: Country of the movie.
    :type country: :class:`~babelfish.country.Country`
    :param list alternative_titles: alternative titles of the movie
    :param \*\*kwargs: additional parameters for the :class:`Video` constructor.

    """
    def __init__(self, name, title, year=None, country=None, alternative_titles=None, **kwargs):
        super(Movie, self).__init__(name, **kwargs)

        #: Title of the movie
        self.title = title

        #: Year of the movie
        self.year = year

        #: Country of the movie
        self.country = country

        #: Alternative titles of the movie
        self.alternative_titles = alternative_titles or []

    def matches(self, title):
        return matches_title(title, self.title, self.alternative_titles)

    @classmethod
    def fromguess(cls, name, guess):
        if guess['type'] != 'movie':
            raise ValueError('The guess must be a movie guess')

        if 'title' not in guess:
            raise ValueError('Insufficient data to process the guess')

        return cls(name, guess['title'], source=guess.get('source'), release_group=guess.get('release_group'),
                   streaming_service=guess.get('streaming_service'),
                   resolution=guess.get('screen_size'), video_codec=guess.get('video_codec'),
                   alternative_titles=ensure_list(guess.get('alternative_title')),
                   audio_codec=guess.get('audio_codec'), year=guess.get('year'), country=guess.get('country'))

    @classmethod
    def fromname(cls, name):
        return cls.fromguess(name, guessit(name, {'type': 'movie'}))

    def __repr__(self):
        return '<{cn} [{title}{open}{country}{sep}{year}{close}]>'.format(
            cn=self.__class__.__name__, title=self.title, year=self.year or '', country=self.country or '',
            open=' (' if self.year or self.country else '',
            sep=') (' if self.year and self.country else '',
            close=')' if self.year or self.country else ''
        )



import platform
import re
import subprocess

from sickchill import logger, settings
from sickchill.oldbeard import helpers, notifiers

from .abstract import UpdateManagerBase


class GitUpdateManager(UpdateManagerBase):
    def __init__(self):
        self._git_path = self._find_working_git()

        self.branch = settings.BRANCH = self._find_installed_branch()

        self._check_detached_head()

        self._cur_commit_hash = None
        self._newest_commit_hash = None
        self._num_commits_behind = 0
        self._num_commits_ahead = 0

    def get_current_commit_hash(self):
        return self._cur_commit_hash

    def get_newest_commit_hash(self):
        return self._newest_commit_hash

    def get_current_version(self):
        return self._run_git(self._git_path, "describe --abbrev=0 {0}".format(self._cur_commit_hash))[0]

    def get_newest_version(self):
        if self._newest_commit_hash:
            return self._run_git(self._git_path, "describe --abbrev=0 " + self._newest_commit_hash)[0]
        else:
            return self._run_git(self._git_path, "describe --abbrev=0 " + self._cur_commit_hash)[0]

    def get_num_commits_behind(self):
        return self._num_commits_behind

    def _find_working_git(self):
        test_cmd = "version"

        if settings.GIT_PATH:
            main_git = '"' + settings.GIT_PATH + '"'
        else:
            main_git = "git"

        logger.debug("Checking if we can use git commands: " + main_git + " " + test_cmd)
        stdout_, stderr_, exit_status = self._run_git(main_git, test_cmd)

        if exit_status == 0:
            logger.debug("Using: " + main_git)
            return main_git
        else:
            logger.debug("Not using: " + main_git)

        # trying alternatives

        alternative_git = []

        # osx people who start sc from launchd have a broken path, so try a hail-mary attempt for them
        if platform.system() == "Darwin":
            alternative_git.append("/usr/local/git/bin/git")

        if platform.system() == "Windows":
            if main_git != main_git.lower():
                alternative_git.append(main_git.lower())

        if alternative_git:
            logger.debug("Trying known alternative git locations")

            for cur_git in alternative_git:
                logger.debug("Checking if we can use git commands: " + cur_git + " " + test_cmd)
                stdout_, stderr_, exit_status = self._run_git(cur_git, test_cmd)

                if exit_status == 0:
                    logger.debug("Using: " + cur_git)
                    return cur_git
                else:
                    logger.debug("Not using: " + cur_git)

        # Still haven't found a working git
        helpers.add_site_message(
            _(
                "Unable to find your git executable - Shutdown SickChill and EITHER set git_path in "
                "your config.ini OR delete your .git folder and run from source to enable updates."
            ),
            tag="unable_to_find_git",
        )
        return None

    @staticmethod
    def _run_git(git_path, args, log_errors=False):

        output = err = exit_status = None

        if not git_path:
            logger.warning("No git specified, can't use git commands")
            exit_status = 1
            return output, err, exit_status

        cmd = git_path + " " + args

        try:
            logger.debug("Executing {0} with your shell in {1}".format(cmd, settings.PROG_DIR))
            p = subprocess.Popen(
                cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True, universal_newlines=True, cwd=settings.PROG_DIR
            )
            output, err = p.communicate()
            exit_status = p.returncode
            if output:
                output = output.strip()

        except OSError:
            logger.info("Command {} didn't work".format(cmd))
            exit_status = 1

        if exit_status == 0:
            logger.debug("{} : returned successful".format(cmd))

        elif exit_status == 1:
            if "stash" in output:
                logger.warning("Please enable 'git reset' in settings or stash your changes in local files")
            elif log_errors:
                logger.exception("{0} returned : {1}".format(cmd, output))

        elif log_errors:
            if exit_status in (127, 128) or "fatal:" in output:
                logger.warning("{0} returned : ({1}) {2}".format(cmd, exit_status, output or err))
            else:
                logger.exception("{0} returned code {1}, treating as error : {2}".format(cmd, exit_status, output or err))
                exit_status = 1

        return output, err, exit_status

    def _find_installed_version(self):
        """
        Attempts to find the currently installed version of SickChill.

        Uses git show to get commit version.

        Returns: True for success or False for failure
        """

        output, errors_, exit_status = self._run_git(self._git_path, "rev-parse HEAD")

        if exit_status == 0 and output:
            cur_commit_hash = output.strip()
            if not re.match("^[a-z0-9]+$", cur_commit_hash):
                logger.exception("Output doesn't look like a hash, not using it")
                return False
            self._cur_commit_hash = settings.CUR_COMMIT_HASH = cur_commit_hash
            return True
        else:
            return False

    def _find_installed_branch(self):
        branch_info, errors_, exit_status = self._run_git(self._git_path, "symbolic-ref -q HEAD")
        if exit_status == 0 and branch_info:
            branch = branch_info.strip().replace("refs/heads/", "", 1)
            if branch:
                settings.BRANCH = branch
                return branch
        return ""

    def _check_detached_head(self):
        # stdout, stderr_, exit_status = self._run_git(self._git_path, 'branch --show-current')
        # if exit_status == 0 and not stdout:
        if helpers.is_docker() and not self.branch:
            logger.info("We found you in a detached state that prevents updates. Fixing")
            self._run_git(self._git_path, "fetch origin")
            stdout_, stderr_, exit_status = self._run_git(self._git_path, "checkout -f master")
            if exit_status == 0:
                self.branch = settings.BRANCH = "master"

    def _check_github_for_update(self):
        """
        Uses git commands to check if there is a newer version that the provided
        commit hash. If there is a newer version it sets _num_commits_behind.
        """

        self._num_commits_behind = 0
        self._num_commits_ahead = 0

        # update remote origin url
        self._update_remote_origin()

        # get all new info from github
        output, errors_, exit_status = self._run_git(self._git_path, "fetch {0} --prune".format(settings.GIT_REMOTE))
        if exit_status != 0:
            logger.warning("Unable to contact github, can't check for update")
            return

        # Try both formats, but continue on fail because older git versions do not have this option
        output, stderr_, exit_status = self._run_git(self._git_path, "branch --set-upstream-to {0}/{1}".format(settings.GIT_REMOTE, self.branch))
        if exit_status != 0:
            self._run_git(self._git_path, "branch -u {0}/{1}".format(settings.GIT_REMOTE, self.branch))

        # get latest commit_hash from remote
        output, stderr_, exit_status = self._run_git(self._git_path, 'rev-parse --verify --quiet "@{upstream}"')

        if exit_status == 0 and output:
            cur_commit_hash = output.strip()

            if not re.match(r"^[a-z0-9]+$", cur_commit_hash):
                logger.debug("Output doesn't look like a hash, not using it")
                return

            else:
                self._newest_commit_hash = cur_commit_hash
        else:
            logger.debug("git didn't return newest commit hash")
            return

        # get number of commits behind and ahead (option --count not supported git < 1.7.2)
        output, stderr_, exit_status = self._run_git(self._git_path, 'rev-list --left-right "@{upstream}"...HEAD')
        if exit_status == 0 and output:

            try:
                self._num_commits_behind = int(output.count("<"))
                self._num_commits_ahead = int(output.count(">"))

            except Exception:
                logger.debug("git didn't return numbers for behind and ahead, not using it")
                return

        logger.debug(
            "cur_commit = {0}, newest_commit = {1}, num_commits_behind = {2}, num_commits_ahead = {3}".format(
                self._cur_commit_hash, self._newest_commit_hash, self._num_commits_behind, self._num_commits_ahead
            )
        )

    def set_newest_text(self):
        if self._num_commits_ahead:
            newest_tag = "local_branch_ahead"
            newest_text = "Local branch is ahead of {branch}. Automatic update not possible.".format(branch=self.branch)
            logger.warning(newest_text)

        elif self._num_commits_behind > 0:
            if self._newest_commit_hash:
                current = self._cur_commit_hash
                newest = self._newest_commit_hash
                url = f"https://github.com/{settings.GIT_ORG}/{settings.GIT_REPO}/compare/{current}...{newest}"
            else:
                url = f"https://github.com/{settings.GIT_ORG}/{settings.GIT_REPO}/commits/"

            newest_tag = "newer_version_available"
            commits_behind = self._num_commits_behind
            s = ("", "s")[commits_behind != 1]
            update_url = self.get_update_url()
            newest_text = _(
                'There is a <a href="{url}" onclick="window.open(this.href); return false;">newer version available</a> (you\'re {commits_behind} commit{s} behind) &mdash; <a href="{update_url}">Update Now</a>'.format(
                    commits_behind=commits_behind, update_url=update_url, url=url, s=s
                )
            )
        else:
            return

        helpers.add_site_message(newest_text, tag=newest_tag, level="success")

    def need_update(self):

        if self.branch != self._find_installed_branch():
            logger.debug("Branch checkout: " + self._find_installed_branch() + "->" + self.branch)
            return True

        self._find_installed_version()
        if not self._cur_commit_hash:
            return True
        else:
            import traceback

            try:
                self._check_github_for_update()
            except Exception as e:
                logger.debug(traceback.format_exc())
                logger.warning("Unable to contact github, can't check for update: " + repr(e))
                return False

            if self._num_commits_behind > 0:
                return True

        return False

    def update(self):
        """
        Calls git pull origin <branch> in order to update SickChill. Returns a bool depending
        on the call's success.
        """

        # update remote origin url
        self._update_remote_origin()

        # remove untracked files and performs a hard reset on git branch to avoid update issues
        if settings.GIT_RESET:
            # self._clean() # This is removing user data and backups
            self._reset()

        if self.branch == self._find_installed_branch():
            stdout_, stderr_, exit_status = self._run_git(self._git_path, "pull -f {0} {1}".format(settings.GIT_REMOTE, self.branch))
        else:
            stdout_, stderr_, exit_status = self._run_git(self._git_path, "checkout -f " + self.branch)

        if exit_status == 0:
            self._find_installed_version()
            self._clean_libs()

            # Notify update successful
            notifiers.notify_git_update(settings.CUR_COMMIT_HASH or "")
            return True
        else:
            return False

    def _clean(self):
        """
        Calls git clean to remove all untracked files. Returns a bool depending
        on the call's success.
        """
        stdout_, stderr_, exit_status = self._run_git(self._git_path, 'clean -df ""')
        if exit_status == 0:
            return True

    def _clean_libs(self):
        """
        Calls git clean to remove all untracked files in the lib dir before restart. Returns a bool depending
        on the call's success.
        """
        stdout_, stderr_, exit_status = self._run_git(self._git_path, "clean -df lib")
        if exit_status == 0:
            return True

        self._clean_pyc("lib3")

    def _reset(self):
        """
        Calls git reset --hard to perform a hard reset. Returns a bool depending
        on the call's success.
        """
        stdout_, stderr_, exit_status = self._run_git(self._git_path, "reset --hard")
        if exit_status == 0:
            return True

    def list_remote_branches(self):
        # update remote origin url
        self._update_remote_origin()
        settings.BRANCH = self._find_installed_branch()

        branches, stderr_, exit_status = self._run_git(self._git_path, "ls-remote --heads {0}".format(settings.GIT_REMOTE))
        if exit_status == 0 and branches:
            if branches:
                return re.findall(r"refs/heads/(.*)", branches)
        return []

    def _update_remote_origin(self):
        if not settings.DEVELOPER:
            self._run_git(self._git_path, "config remote.{0}.url {1}".format(settings.GIT_REMOTE, settings.GIT_REMOTE_URL))



import html
import json
import re
import traceback

import requests
import tvdbsimple

# from sickchill import logger
import sickchill.start
from sickchill import logger, settings
from sickchill.tv import TVEpisode

from .base import Indexer
from .wrappers import ExceptionDecorator


class TVDB(Indexer):
    def __init__(self):
        super(TVDB, self).__init__()
        self.name = "theTVDB"
        self.slug = "tvdb"
        self.api_key = "6aa6e4ecae5b56e9644f6a303c0739b6"
        self.show_url = "https://thetvdb.com/?tab=series&id="
        self.base_url = "https://thetvdb.com/api/%(apikey)s/series/"
        self.icon = "images/indexers/thetvdb16.png"
        tvdbsimple.keys.API_KEY = self.api_key
        self._search = tvdbsimple.search.Search().series
        self._series = tvdbsimple.Series
        self.series_episodes = tvdbsimple.Series_Episodes
        self.series_images = tvdbsimple.Series_Images
        self.updates = tvdbsimple.Updates

    @ExceptionDecorator()
    def series(self, *args, **kwargs):
        result = self._series(*args, **kwargs)
        if result:
            result.info(language=kwargs.get("language"))
        return result

    @ExceptionDecorator()
    def get_series_by_id(self, indexerid, language=None):
        result = self._series(indexerid, language)
        if result:
            result.info(language=language)
        return result

    @ExceptionDecorator()
    def series_from_show(self, show):
        result = self._series(show.indexerid, show.lang)
        if result:
            result.info(language=show.lang)
        return result

    def series_from_episode(self, episode):
        return self.series_from_show(episode.show)

    def get_series_by_name(self, name, indexerid=None, language=None):
        if indexerid:
            return self.get_series_by_id(indexerid, language)

        # Just return the first result for now
        result = self._series(self.search(name, language)[0]["id"])
        if result:
            result.info(language=language)
        return result

    @ExceptionDecorator()
    def episodes(self, show, season=None):
        if show.dvdorder:
            result = self.series_episodes(show.indexerid, dvdSeason=season, language=show.lang).all()
        else:
            result = self.series_episodes(show.indexerid, airedSeason=season, language=show.lang).all()

        return result

    @ExceptionDecorator()
    def episode(self, item, season=None, episode=None, **kwargs):
        if isinstance(item, TVEpisode):
            show = item.show
            season = item.season
            episode = item.episode
        else:
            show = item

        if show.dvdorder:
            result = self.series_episodes(show.indexerid, dvdSeason=season, dvdEpisode=episode, language=show.lang, **kwargs).all()[0]
        else:
            result = self.series_episodes(show.indexerid, airedSeason=season, airedEpisode=episode, language=show.lang, **kwargs).all()[0]

        return result

    @ExceptionDecorator(default_return=list())
    def search(self, name, language=None, exact=False, indexer_id=False):
        """
        :param name: Show name to search for
        :param language: Language of the show info we want
        :param exact: Exact when adding existing, processed when adding new shows
        :param indexer_id: Exact indexer id to get, either imdb or tvdb id.
        :return: list of series objects
        """
        language = language or self.language
        result = []
        if isinstance(name, bytes):
            name = name.decode()

        if re.match(r"^t?t?\d{7,8}$", name) or re.match(r"^\d{6}$", name):
            try:
                if re.match(r"^t?t?\d{7,8}$", name):
                    result = self._search(imdbId=f'tt{name.strip("t")}', language=language)
                elif re.match(r"^\d{6}$", name):
                    series = self._series(name, language=language)
                    if series:
                        result = [series.info(language)]
            except requests.exceptions.RequestException:
                logger.exception(traceback.format_exc())
        else:
            # Name as provided (usually from nfo)
            names = [name]
            if not exact:
                # Name without year and separator
                test = re.match(r"^(.+?)[. -]+\(\d{4}\)?$", name)
                if test:
                    names.append(test.group(1).strip())
                # Name with spaces
                if re.match(r"[. -_]", name):
                    names.append(re.sub(r"[. -_]", " ", name).strip())
                    if test:
                        # Name with spaces and without year
                        names.append(re.sub(r"[. -_]", " ", test.group(1)).strip())

            for attempt in set(n for n in names if n.strip()):
                try:
                    result = self._search(attempt, language=language)
                    if result:
                        break
                except requests.exceptions.RequestException:
                    logger.exception(traceback.format_exc())

        return result

    @property
    def languages(self):
        return ["da", "fi", "nl", "de", "it", "es", "fr", "pl", "hu", "el", "tr", "ru", "he", "ja", "pt", "zh", "cs", "sl", "hr", "ko", "en", "sv", "no"]

    @property
    def lang_dict(self):
        return {
            "el": 20,
            "en": 7,
            "zh": 27,
            "it": 15,
            "cs": 28,
            "es": 16,
            "ru": 22,
            "nl": 13,
            "pt": 26,
            "no": 9,
            "tr": 21,
            "pl": 18,
            "fr": 17,
            "hr": 31,
            "de": 14,
            "da": 10,
            "fi": 11,
            "hu": 19,
            "ja": 25,
            "he": 24,
            "ko": 32,
            "sv": 8,
            "sl": 30,
        }

    @staticmethod
    def complete_image_url(location):
        location = location.strip()
        if not location:
            return location
        return f'https://artworks.thetvdb.com/banners/{re.sub(r"^_cache/", "", location)}'

    @ExceptionDecorator(default_return="", catch=(requests.exceptions.RequestException, KeyError), image_api=True)
    def __call_images_api(self, show, thumb, keyType, subKey=None, lang=None, multiple=False):
        api_results = self.series_images(show.indexerid, lang or show.lang, keyType=keyType, subKey=subKey)
        images = getattr(api_results, keyType)(lang or show.lang)
        images = sorted(images, key=lambda img: img["ratingsInfo"]["average"], reverse=True)
        return [self.complete_image_url(image["fileName"]) for image in images] if multiple else self.complete_image_url(images[0]["fileName"])

    @staticmethod
    @ExceptionDecorator()
    def actors(series):
        if hasattr(series, "actors") and callable(series.actors):
            series.actors(series.language)
        return series.actors

    def series_poster_url(self, show, thumb=False, multiple=False):
        return self.__call_images_api(show, thumb, "poster", multiple=multiple)

    def series_banner_url(self, show, thumb=False, multiple=False):
        return self.__call_images_api(show, thumb, "series", multiple=multiple)

    def series_fanart_url(self, show, thumb=False, multiple=False):
        return self.__call_images_api(show, thumb, "fanart", multiple=multiple)

    def season_poster_url(self, show, season, thumb=False, multiple=False):
        return self.__call_images_api(show, thumb, "season", season, multiple=multiple)

    def season_banner_url(self, show, season, thumb=False, multiple=False):
        return self.__call_images_api(show, thumb, "seasonwide", season, multiple=multiple)

    @ExceptionDecorator(default_return="", catch=(requests.exceptions.RequestException, KeyError, TypeError))
    def episode_image_url(self, episode):
        return self.complete_image_url(self.episode(episode)["filename"])

    def episode_guide_url(self, show):
        # https://forum.kodi.tv/showthread.php?tid=323588
        data = html.escape(json.dumps({"apikey": self.api_key, "id": show.indexerid})).replace(" ", "")
        return tvdbsimple.base.TVDB(key=self.api_key)._get_complete_url("login") + "?" + data + "|Content-Type=application/json"

    def get_favorites(self):
        results = []
        if not (settings.TVDB_USER and settings.TVDB_USER_KEY):
            return results

        user = tvdbsimple.User(settings.TVDB_USER, settings.TVDB_USER_KEY)
        for tvdbid in user.favorites():
            results.append(self.get_series_by_id(tvdbid))

        return results

    @staticmethod
    def test_user_key(user, key):
        user_object = tvdbsimple.User(user, key)
        try:
            user_object.info()
        except Exception:
            logger.exception(traceback.format_exc())
            return False

        settings.TVDB_USER = user
        settings.TVDB_USER_KEY = key

        sickchill.start.save_config()

        return True



# Copyright: 2013 Paul Traylor
# These sources are released under the terms of the MIT license: see LICENSE

class BaseError(Exception):
	pass


class ParseError(BaseError):
	errorcode = 500
	errordesc = 'Error parsing the message'


class AuthError(BaseError):
	errorcode = 400
	errordesc = 'Error with authorization'


class UnsupportedError(BaseError):
	errorcode = 500
	errordesc = 'Currently unsupported by gntp.py'


class NetworkError(BaseError):
	errorcode = 500
	errordesc = "Error connecting to growl server"



import re
from fnmatch import fnmatch

from github import Github
from github.GithubException import (
    BadAttributeException,
    BadCredentialsException,
    BadUserAgentException,
    GithubException,
    RateLimitExceededException,
    TwoFactorException,
    UnknownObjectException,
)

import sickchill
from sickchill import settings

dateFormat = "%Y-%m-%d"
dateTimeFormat = "%Y-%m-%d %H:%M:%S"

# Mapping HTTP status codes to official W3C names
HTTP_STATUS_CODES = {
    300: "Multiple Choices",
    301: "Moved Permanently",
    302: "Found",
    303: "See Other",
    304: "Not Modified",
    305: "Use Proxy",
    306: "Switch Proxy",
    307: "Temporary Redirect",
    308: "Permanent Redirect",
    400: "Bad Request",
    401: "Unauthorized",
    402: "Payment Required",
    403: "Forbidden",
    404: "Not Found",
    405: "Method Not Allowed",
    406: "Not Acceptable",
    407: "Proxy Authentication Required",
    408: "Request Timeout",
    409: "Conflict",
    410: "Gone",
    411: "Length Required",
    412: "Precondition Failed",
    413: "Request Entity Too Large",
    414: "Request-URI Too Long",
    415: "Unsupported Media Type",
    416: "Requested Range Not Satisfiable",
    417: "Expectation Failed",
    418: "Im a teapot",
    419: "Authentication Timeout",
    420: "Enhance Your Calm",
    422: "Unprocessable Entity",
    423: "Locked",
    424: "Failed Dependency",
    426: "Upgrade Required",
    428: "Precondition Required",
    429: "Too Many Requests",
    431: "Request Header Fields Too Large",
    440: "Login Timeout",
    444: "No Response",
    449: "Retry With",
    450: "Blocked by Windows Parental Controls",
    451: [
        "Redirect",
        "Unavailable For Legal Reasons",
    ],
    494: "Request Header Too Large",
    495: "Cert Error",
    496: "No Cert",
    497: "HTTP to HTTPS",
    498: "Token expired/invalid",
    499: [
        "Client Closed Request",
        "Token required",
    ],
    500: "Internal Server Error",
    501: "Not Implemented",
    502: "Bad Gateway",
    503: "Service Unavailable",
    504: "Gateway Timeout",
    505: "HTTP Version Not Supported",
    506: "Variant Also Negotiates",
    507: "Insufficient Storage",
    508: "Loop Detected",
    509: "Bandwidth Limit Exceeded",
    510: "Not Extended",
    511: "Network Authentication Required",
    520: "CloudFlare - Web server is returning an unknown error",
    521: "CloudFlare - Web server is down",
    522: "CloudFlare - Connection timed out",
    523: "CloudFlare - Origin is unreachable",
    524: "CloudFlare - A timeout occurred",
    525: "CloudFlare - SSL handshake failed",
    526: "CloudFlare - Invalid SSL certificate",
    598: "Network read timeout error",
    599: "Network connect timeout error",
}
MEDIA_EXTENSIONS = [
    "3gp",
    "avi",
    "divx",
    "dvr-ms",
    "f4v",
    "flv",
    "m2ts",
    "m4v",
    "mkv",
    "mov",
    "mp4",
    "mpeg",
    "mpg",
    "ogm",
    "ogv",
    "rmvb",
    "tp",
    "ts",
    "vob",
    "webm",
    "wmv",
    "wtv",
]

SUBTITLE_EXTENSIONS = ["ass", "idx", "srt", "ssa", "sub"]
timeFormat = "%A %I:%M %p"


def http_code_description(http_code):
    """
    Get the description of the provided HTTP status code.
    :param http_code: The HTTP status code
    :return: The description of the provided ``http_code``
    """

    description = HTTP_STATUS_CODES.get(try_int(http_code))

    if isinstance(description, list):
        return "({0})".format(", ".join(description))

    return description


def is_sync_file(filename):
    """
    Check if the provided ``filename`` is a sync file, based on its name.
    :param filename: The filename to check
    :return: ``True`` if the ``filename`` is a sync file, ``False`` otherwise
    """

    if isinstance(filename, str):
        extension = filename.rpartition(".")[2].lower()

        return (
            extension in settings.SYNC_FILES.split(",")
            or filename.startswith(".syncthing")
            or any(fnmatch(filename, match) for match in settings.SYNC_FILES.split(","))
        )

    return False


def is_torrent_or_nzb_file(filename):
    """
    Check if the provided ``filename`` is a NZB file or a torrent file, based on its extension.
    :param filename: The filename to check
    :return: ``True`` if the ``filename`` is a NZB file or a torrent file, ``False`` otherwise
    """

    if not isinstance(filename, str):
        return False

    return filename.rpartition(".")[2].lower() in ["nzb", "torrent"]


def pretty_file_size(size, use_decimal=False, **kwargs):
    """
    Return a human readable representation of the provided ``size``.

    :param size: The size to convert
    :param use_decimal: use decimal instead of binary prefixes (e.g. kilo = 1000 instead of 1024)

    :keyword units: A list of unit names in ascending order.
        Default units: ['B', 'KB', 'MB', 'GB', 'TB', 'PB']

    :return: The converted size
    """
    try:
        size = max(float(size), 0.0)
    except (ValueError, TypeError):
        size = 0.0

    remaining_size = size
    units = kwargs.pop("units", ["B", "KB", "MB", "GB", "TB", "PB"])
    block = 1024.0 if not use_decimal else 1000.0
    for unit in units:
        if remaining_size < block:
            return "{0:3.2f} {1}".format(remaining_size, unit)
        remaining_size /= block
    return size


def convert_size(size, default=None, use_decimal=False, **kwargs):
    """
    Convert a file size into the number of bytes

    :param size: to be converted
    :param default: value to return if conversion fails
    :param use_decimal: use decimal instead of binary prefixes (e.g. kilo = 1000 instead of 1024)

    :keyword sep: Separator between size and units, default is space
    :keyword units: A list of (uppercase) unit names in ascending order.
        Default units: ['B', 'KB', 'MB', 'GB', 'TB', 'PB']

    :keyword default_units: Default unit if none is given,
        default is lowest unit on the scale, e.g. bytes

    :returns: the number of bytes, the default value, or 0
    """
    result = None

    try:
        sep = kwargs.pop("sep", " ")
        scale = kwargs.pop("units", ["B", "KB", "MB", "GB", "TB", "PB"])
        default_units = kwargs.pop("default_units", scale[0])

        if sep:
            size_tuple = size.strip().split(sep)
            scalar, units = size_tuple[0], size_tuple[1:]
            units = units[0].upper() if units else default_units
        else:
            regex_scalar = re.search(r"([\d. ]+)", size, re.I)
            scalar = regex_scalar.group() if regex_scalar else -1
            units = size.strip(scalar) if scalar != -1 else "B"

        scalar = float(scalar)
        scalar *= (1024 if not use_decimal else 1000) ** scale.index(units)

        result = scalar

    # TODO: Make sure fallback methods obey default units
    except AttributeError:
        result = size if size is not None else default

    except ValueError:
        result = default

    finally:
        try:
            if result != default:
                result = max(int(result), 0)
        except (TypeError, ValueError):
            pass

    return result


def remove_extension(filename):
    """
    Remove the extension of the provided ``filename``.
    The extension is only removed if it is in MEDIA_EXTENSIONS or ['nzb', 'torrent'].
    :param filename: The filename from which we want to remove the extension
    :return: The ``filename`` without its extension.
    """

    if isinstance(filename, str) and "." in filename:
        basename, dot, extension = filename.rpartition(".")

        if basename and extension.lower() in ["nzb", "torrent"] + MEDIA_EXTENSIONS:
            return basename

    return filename


def replace_extension(filename, new_extension):
    """
    Replace the extension of the provided ``filename`` with a new extension.
    :param filename: The filename for which we want to change the extension
    :param new_extension: The new extension to apply on the ``filename``
    :return: The ``filename`` with the new extension
    """

    if isinstance(filename, str) and "." in filename:
        basename = filename.rpartition(".")[0]
        if basename:
            return "{0}.{1}".format(basename, new_extension)

    return filename


def sanitize_filename(filename):
    """
    Remove specific characters from the provided ``filename``.
    :param filename: The filename to clean
    :return: The ``filename``cleaned
    """

    if isinstance(filename, bytes):
        filename = filename.decode()

    if isinstance(filename, str):
        filename = re.sub(r"[\\/*]", "-", filename)
        filename = re.sub(r'[:"<>|?]', "", filename)
        filename = re.sub(r"™|-u2122", "", filename)  # Trade Mark Sign unicode: \u2122
        filename = filename.strip(" .")

        return filename

    return ""


def try_int(candidate, default_value=0):
    """
    Try to convert ``candidate`` to int, or return the ``default_value``.
    :param candidate: The value to convert to int
    :param default_value: The value to return if the conversion fails
    :return: ``candidate`` as int, or ``default_value`` if the conversion fails
    """

    try:
        return int(candidate)
    except (ValueError, TypeError):
        return default_value


def episode_num(season=None, episode=None, **kwargs):
    """
    Convert season and episode into string

    :param season: Season number
    :param episode: Episode Number
    :keyword numbering: Absolute for absolute numbering
    :returns: a string in s01e01 format or absolute numbering
    """

    numbering = kwargs.pop("numbering", "standard")

    if numbering == "standard":
        if season is not None and episode:
            return "S{0:0>2}E{1:02}".format(season, episode)
    elif numbering == "absolute":
        if not (season and episode) and (season or episode):
            return "{0:0>3}".format(season or episode)


def setup_github():
    """
    Instantiate the global github connection, for checking for updates and submitting issues
    """

    try:
        if settings.GIT_TOKEN:
            # Token Auth - allows users with Two-Factor Authorization (2FA) enabled on Github to connect their account.
            settings.gh = Github(login_or_token=settings.GIT_TOKEN, user_agent="SickChill")
            # This will trigger:
            # * BadCredentialsException if token is invalid
            # * TwoFactorException if user has enabled Github-2FA
            #   but didn't set a personal token in the configuration.
            settings.gh.get_organization(settings.GIT_ORG)
        if not settings.gh:
            settings.gh = Github(user_agent="SickChill")
            settings.gh.get_organization(settings.GIT_ORG)
    except BadCredentialsException as error:
        settings.gh = None
        sickchill.logger.warning(_("Unable to setup GitHub properly with your github token. Please check your credentials. Error: {0}").format(error))
    except TwoFactorException as error:
        settings.gh = None
        sickchill.logger.warning(
            _("Unable to setup GitHub properly with your github token due to 2FA - Make sure this token works with 2FA. Error: {0}").format(error)
        )
    except RateLimitExceededException as error:
        settings.gh = None
        if settings.GIT_TOKEN:
            sickchill.logger.warning(
                _("Unable to setup GitHub properly, You are currently being throttled by rate limiting for too many requests. Error: {0}").format(error)
            )
        else:
            sickchill.logger.warning(
                _(
                    "Unable to setup GitHub properly, You are currently being throttled by rate limiting for too many requests - Try adding an access token. Error: {0}"
                ).format(error)
            )
    except UnknownObjectException as error:
        settings.gh = None
        sickchill.logger.warning(_("Unable to setup GitHub properly, it seems to be down or your organization/repo is set wrong. Error: {0}").format(error))
    except BadUserAgentException as error:
        settings.gh = None
        sickchill.logger.warning(_("Unable to setup GitHub properly, GitHub doesn't like the user-agent. Error: {0}").format(error))
    except BadAttributeException as error:
        settings.gh = None
        sickchill.logger.error(_("Unable to setup GitHub properly, There might be an error with the library. Error: {0}").format(error))
    except (GithubException, Exception) as error:
        settings.gh = None
        sickchill.logger.error(_("Unable to setup GitHub properly. GitHub will not be available. Error: {0}").format(error))



import re
from urllib.parse import urljoin

from requests.utils import dict_from_cookiejar

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    """BJ-Share Torrent provider."""

    def __init__(self):
        """Initialize the class."""
        super().__init__("BJ-Share")

        # URLs
        self.url = "https://bj-share.info"
        self.urls = {"login": "https://bj-share.info/login.php", "search": urljoin(self.url, "torrents.php")}

        # Credentials
        self.enable_cookies = True
        self.cookies = ""
        self.username = None
        self.password = None
        self.required_cookies = ["session"]

        # Torrent Stats
        self.minseed = 0
        self.minleech = 0

        # Miscellaneous Options
        self.supports_absolute_numbering = True
        self.max_back_pages = 2

        # Proper Strings
        self.proper_strings = ["PROPER", "REPACK", "REAL", "RERIP"]

        # Cache
        self.cache = tvcache.TVCache(self)

        # One piece and Boruto is the only animes that i'm aware that is in "absolute" numbering, the problem is that
        # they include the season (wrong season) and episode as absolute, eg: One Piece - S08E836
        # 836 is the latest episode in absolute numbering, that is correct, but S08 is not the current season...
        # So for this show, i don't see a other way to make it work...
        #
        # All others animes that i tested is with correct season and episode set, so i can't remove the season from all
        # or will break everything else
        #
        # In this indexer, it looks that it is added "automatically", so all current and new releases will be broken
        # until they or the source from where they get that info fix it...
        self.absolute_numbering = ["One Piece", "Boruto: Naruto Next Generations"]

    def search(self, search_strings, age=0, ep_obj=None):
        """
        Search a provider and parse the results.

        :param search_strings: A dict with mode (key) and the search value (value)
        :param age: Not used
        :param ep_obj: Informations about the episode being searched (when not RSS)

        :returns: A list of search results (structure)
        """
        results = []
        if not self.login():
            return results

        anime = False
        if ep_obj and ep_obj.show:
            anime = ep_obj.show.anime == 1

        search_params = {"order_by": "time", "order_way": "desc", "group_results": 0, "action": "basic", "searchsubmit": 1}

        if "RSS" in list(search_strings):
            search_params["filter_cat[14]"] = 1  # anime
            search_params["filter_cat[2]"] = 1  # tv shows
        elif anime:
            search_params["filter_cat[14]"] = 1  # anime
        else:
            search_params["filter_cat[2]"] = 1  # tv shows

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))

            # if looking for season, look for more pages
            if mode == "Season":
                self.max_back_pages = 10

            for search_string in {*search_strings[mode]}:
                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))

                # Remove season / episode from search (not supported by tracker)
                search_str = re.sub(r"\d+$" if anime else r"[S|E]\d\d", "", search_string).strip()
                search_params["searchstr"] = search_str
                next_page = 1
                has_next_page = True

                while has_next_page and next_page <= self.max_back_pages:
                    search_params["page"] = next_page
                    logger.debug("Page Search: {0}".format(next_page))
                    next_page += 1

                    response = self.session.get(self.urls["search"], params=search_params)
                    if not response:
                        logger.debug(_("No data returned from provider"))
                        continue

                    result = self._parse(response.content, mode)
                    has_next_page = result["has_next_page"]
                    items += result["items"]

                results += items

        return results

    def _parse(self, data, mode):
        """
        Parse search results for items.

        :param data: The raw response from a search
        :param mode: The current mode used to search, e.g. RSS

        :return: A KV with a list of items found and if there's an next page to search
        """

        def process_column_header(td):
            ret = ""
            if td.a and td.a.img:
                ret = td.a.img.get("title", td.a.get_text(strip=True))
            if not ret:
                ret = td.get_text(strip=True)
            return ret

        items = []
        has_next_page = False
        with BS4Parser(data, "html5lib") as html:
            torrent_table = html.find("table", id="torrent_table")
            torrent_rows = torrent_table("tr") if torrent_table else []

            # ignore next page in RSS mode
            has_next_page = mode != "RSS" and html.find("a", class_="pager_next") is not None
            logger.debug("More Pages? {0}".format(has_next_page))

            # Continue only if at least one Release is found
            if len(torrent_rows) < 2:
                logger.debug("Data returned from provider does not contain any torrents")
                return {"has_next_page": has_next_page, "items": []}

            # '', '', 'Name /Year', 'Files', 'Time', 'Size', 'Snatches', 'Seeders', 'Leechers'
            labels = [process_column_header(label) for label in torrent_rows[0]("td")]
            group_title = ""

            # Skip column headers
            for result in torrent_rows[1:]:
                cells = result("td")
                result_class = result.get("class")
                # When "Grouping Torrents" is enabled, the structure of table change
                group_index = -2 if "group_torrent" in result_class else 0
                try:
                    title = result.select('a[href^="torrents.php?id="]')[0].get_text()
                    title = re.sub(r"\s+", " ", title).strip()  # clean empty lines and multiple spaces

                    if "group" in result_class or "torrent" in result_class:
                        # get international title if available
                        title = re.sub(r".* \[(.*?)\](.*)", r"\1\2", title)

                    if "group" in result_class:
                        group_title = title
                        continue

                    # Clean dash between title and season/episode
                    title = re.sub(r"- (S\d{2}(E\d{2,4})?)", r"\1", title)

                    for serie in self.absolute_numbering:
                        if serie in title:
                            # remove season from title when its in absolute format
                            title = re.sub(r"S\d{2}E(\d{2,4})", r"\1", title)
                            break

                    download_url = urljoin(self.url, result.select('a[href^="torrents.php?action=download"]')[0]["href"])
                    if not all([title, download_url]):
                        continue

                    # seeders = try_int(cells[labels.index('Seeders') + group_index].get_text(strip=True))
                    # leechers = try_int(cells[labels.index('Leechers') + group_index].get_text(strip=True))

                    seeders = try_int(cells[4].get_text(strip=True))
                    leechers = try_int(cells[5].get_text(strip=True))

                    # Filter unseeded torrent
                    if seeders < self.minseed or leechers < self.minleech:
                        if mode != "RSS":
                            logger.debug(
                                "Discarding torrent because it doesn't meet the"
                                " minimum seeders or leechers: {0} (S:{1} L:{2})".format(title, seeders, leechers)
                            )
                        continue

                    torrent_details = None
                    if "group_torrent" in result_class:
                        # torrents belonging to a group
                        torrent_details = title
                        title = group_title
                    elif "torrent" in result_class:
                        # standalone/un grouped torrents
                        torrent_details = cells[labels.index("Nome/Ano")].find("div", class_="torrent_info").get_text()

                    torrent_details = torrent_details.replace("[", " ").replace("]", " ").replace("/", " ")
                    torrent_details = torrent_details.replace("Full HD ", "1080p").replace("HD ", "720p")

                    # torrent_size = cells[labels.index('Tamanho') + group_index].get_text(strip=True)
                    torrent_size = cells[2].get_text(strip=True)

                    size = convert_size(torrent_size) or -1

                    torrent_name = "{0} {1}".format(title, torrent_details.strip()).strip()
                    torrent_name = re.sub(r"\s+", " ", torrent_name)

                    items.append({"title": torrent_name, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""})

                    if mode != "RSS":
                        logger.debug(
                            _(
                                "Found result: {title} with {seeders} seeders and {leechers} leechers".format(
                                    title=torrent_name, seeders=seeders, leechers=leechers
                                )
                            )
                        )

                except (AttributeError, TypeError, KeyError, ValueError, IndexError):
                    logger.exception("Failed parsing provider.")

        return {"has_next_page": has_next_page, "items": items}

    def login(self):
        """Login method used for logging in before doing a search and torrent downloads."""
        cookie_dict = dict_from_cookiejar(self.session.cookies)
        if cookie_dict.get("session"):
            return True

        if self.cookies:
            self.add_cookies_from_ui()

        cookie_dict = dict_from_cookiejar(self.session.cookies)
        if cookie_dict.get("session"):
            return True

        login_params = {
            "submit": "Login",
            "username": self.username,
            "password": self.password,
            "keeplogged": 1,
        }

        response = self.get_url(self.urls["login"], post_data=login_params, returns="text")
        if not response:
            logger.warning("Unable to connect to provider")
            return False

        if re.search("<title>Login :: BJ-Share</title>", response):
            logger.warning("Invalid username or password. Check your settings")
            return False

        return True



import copy
import io
import logging
import re
from xml.etree import ElementTree
from zipfile import is_zipfile, ZipFile

from babelfish import Language
from guessit import guessit
from requests import Session
from subliminal import __version__
from subliminal.cache import EPISODE_EXPIRATION_TIME, region, SHOW_EXPIRATION_TIME
from subliminal.exceptions import AuthenticationError, ConfigurationError, DownloadLimitExceeded
from subliminal.matches import guess_matches, sanitize
from subliminal.providers import Provider
from subliminal.subtitle import fix_line_ending, Subtitle
from subliminal.video import Episode

logger = logging.getLogger(__name__)


class ItaSASubtitle(Subtitle):
    provider_name = "itasa"

    def __init__(self, sub_id, series, season, episode, source, year, tvdb_id, full_data):
        super().__init__(Language("ita"))
        self.sub_id = sub_id
        self.series = series
        self.season = season
        self.episode = episode
        self.source = source
        self.year = year
        self.tvdb_id = tvdb_id
        self.full_data = full_data

    @property
    def id(self):  # pragma: no cover
        return self.sub_id

    def get_matches(self, video: Episode, hearing_impaired=False):
        matches = set()

        # series
        if video.series and sanitize(self.series) == sanitize(video.series):
            matches.add("series")
        # season
        if video.season and self.season == video.season:
            matches.add("season")
        # episode
        if video.episode and self.episode == video.episode:
            matches.add("episode")
        # source
        if video.source and video.source.lower() in self.source.lower():
            matches.add("source")
        if video.year and self.year == video.year:
            matches.add("year")
        if video.series_tvdb_id and self.tvdb_id == video.series_tvdb_id:
            matches.add("series_tvdb_id")

        # other properties
        matches |= guess_matches(video, guessit(self.full_data), partial=True)

        return matches


class ItaSAProvider(Provider):
    languages = {Language("ita")}

    video_types = (Episode,)

    server_url = "https://api.italiansubs.net/api/rest/"

    apikey = "d86ad6ec041b334fac1e512174ee04d5"

    def __init__(self, username=None, password=None):
        if username is not None and password is None or username is None and password is not None:
            raise ConfigurationError("Username and password must be specified")

        self.username = username
        self.password = password
        self.logged_in = False
        self.login_itasa = False
        self.session = None
        self.auth_code = None

    def initialize(self):
        self.session = Session()
        self.session.headers["User-Agent"] = "Subliminal/%s" % __version__

        # login
        if self.username is not None and self.password is not None:
            logger.info("Logging in")
            params = {"username": self.username, "password": self.password, "apikey": self.apikey}

            r = self.session.get(self.server_url + "users/login", params=params, allow_redirects=False, timeout=10)
            root = ElementTree.fromstring(r.content)

            if root.find("status").text == "fail":
                raise AuthenticationError(root.find("error/message").text)

            self.auth_code = root.find("data/user/authcode").text

            r = self.session.get("https://www.italiansubs.net/", allow_redirects=True, timeout=30)
            form = re.search(r'<form.*?id="form-login".*?>.*?</form>', r.text, re.DOTALL)
            input_matcher = re.finditer('<input.*?name="(.*?)".*?value="(.*?)".*?/>', form.group())

            data = {}

            for inputMatch in input_matcher:
                data[inputMatch.group(1)] = inputMatch.group(2)
            data["username"] = self.username
            data["passwd"] = self.password

            r = self.session.post("https://www.italiansubs.net/index.php", data=data, allow_redirects=True, timeout=30)
            r.raise_for_status()

            self.logged_in = True

    def terminate(self):
        self.session.close()
        self.logged_in = False

    @region.cache_on_arguments(expiration_time=SHOW_EXPIRATION_TIME)
    def _get_show_ids(self):
        """Get the ``dict`` of show ids per series by querying the `shows` page.

        :return: show id per series, lower case and without quotes.
        :rtype: dict

        """
        # get the show page
        logger.info("Getting show ids")
        params = {"apikey": self.apikey}
        r = self.session.get(self.server_url + "shows", timeout=10, params=params)
        r.raise_for_status()
        root = ElementTree.fromstring(r.content)

        # populate the show ids
        show_ids = {}
        for show in root.findall("data/shows/show"):
            if show.find("name").text is None:  # pragma: no cover
                continue
            show_ids[sanitize(show.find("name").text).lower()] = int(show.find("id").text)
        logger.debug("Found %d show ids", len(show_ids))

        return show_ids

    @region.cache_on_arguments(expiration_time=SHOW_EXPIRATION_TIME)
    def _search_show_id(self, series):
        """Search the show id from the `series`

        :param str series: series of the episode.
        :return: the show id, if found.
        :rtype: int or None

        """
        # build the param
        params = {"apikey": self.apikey, "q": series}

        # make the search
        logger.info("Searching show ids with %r", params)
        r = self.session.get(self.server_url + "shows/search", params=params, timeout=10)
        r.raise_for_status()
        root = ElementTree.fromstring(r.content)

        if int(root.find("data/count").text) == 0:
            logger.warning("Show id not found: no suggestion")
            return None

        # Looking for show in first page
        for show in root.findall("data/shows/show"):
            if sanitize(show.find("name").text).lower() == sanitize(series.lower()):
                show_id = int(show.find("id").text)
                logger.debug("Found show id %d", show_id)

                return show_id

        # Not in the first page of result try next (if any)
        next_page = root.find("data/next")
        while next_page.text is not None:  # pragma: no cover

            r = self.session.get(next_page.text, timeout=10)
            r.raise_for_status()
            root = ElementTree.fromstring(r.content)

            logger.info("Loading suggestion page %r", root.find("data/page").text)

            # Looking for show in following pages
            for show in root.findall("data/shows/show"):
                if sanitize(show.find("name").text).lower() == sanitize(series.lower()):
                    show_id = int(show.find("id").text)
                    logger.debug("Found show id %d", show_id)

                    return show_id

            next_page = root.find("data/next")

        # No matches found
        logger.warning("Show id not found: suggestions does not match")

        return None

    def get_show_id(self, series, country_code=None):
        """Get the best matching show id for `series`.

        First search in the result of :meth:`_get_show_ids` and fallback on a search with :meth:`_search_show_id`

        :param str series: series of the episode.
        :param str country_code: the country in which teh show is aired.
        :return: the show id, if found.
        :rtype: int or None

        """
        series_sanitized = sanitize(series).lower()
        show_ids = self._get_show_ids()
        show_id = None

        # attempt with country
        if not show_id and country_code:
            logger.debug("Getting show id with country")
            show_id = show_ids.get("%s %s" % (series_sanitized, country_code.lower()))

        # attempt clean
        if not show_id:
            logger.debug("Getting show id")
            show_id = show_ids.get(series_sanitized)

        # search as last resort
        if not show_id:
            logger.warning("Series not found in show ids")
            show_id = self._search_show_id(series)

        return show_id

    @region.cache_on_arguments(expiration_time=EPISODE_EXPIRATION_TIME)
    def _download_zip(self, sub_id):
        # download the subtitle
        logger.info("Downloading subtitle %r", sub_id)

        params = {"authcode": self.auth_code, "apikey": self.apikey, "subtitle_id": sub_id}

        r = self.session.get(self.server_url + "subtitles/download", params=params, timeout=30)
        r.raise_for_status()

        return r.content

    def _get_season_subtitles(self, show_id, season, sub_source):
        params = {"apikey": self.apikey, "show_id": show_id, "q": "Stagione %%%d" % season, "version": sub_source}
        r = self.session.get(self.server_url + "subtitles/search", params=params, timeout=30)
        r.raise_for_status()
        root = ElementTree.fromstring(r.content)

        if int(root.find("data/count").text) == 0:
            logger.warning("Subtitles for season not found, try with rip suffix")

            params["version"] = sub_source + "rip"
            r = self.session.get(self.server_url + "subtitles/search", params=params, timeout=30)
            r.raise_for_status()
            root = ElementTree.fromstring(r.content)
            if int(root.find("data/count").text) == 0:
                logger.warning("Subtitles for season not found")
                return []

        subs = []
        # Looking for subtitles in first page
        season_re = re.compile(".*?stagione 0*?%d.*" % season)
        for subtitle in root.findall("data/subtitles/subtitle"):
            if season_re.match(subtitle.find("name").text.lower()):
                logger.debug("Found season zip id %d - %r - %r", int(subtitle.find("id").text), subtitle.find("name").text, subtitle.find("version").text)

                content = self._download_zip(int(subtitle.find("id").text))
                if not is_zipfile(io.BytesIO(content)):  # pragma: no cover
                    if "limite di download" in content:
                        raise DownloadLimitExceeded()
                    else:
                        raise ConfigurationError("Not a zip file: %r" % content)

                with ZipFile(io.BytesIO(content)) as zf:
                    episode_re = re.compile(r"s(\d{1,2})e(\d{1,2})")
                    for index, name in enumerate(zf.namelist()):
                        match = episode_re.search(name)
                        if not match:  # pragma: no cover
                            logger.debug("Cannot decode subtitle %r", name)
                        else:
                            sub = ItaSASubtitle(
                                int(subtitle.find("id").text), subtitle.find("show_name").text, int(match.group(1)), int(match.group(2)), None, None, None, name
                            )
                            sub.content = fix_line_ending(zf.read(name))
                            subs.append(sub)

        return subs

    def query(self, series, season, episode, source, resolution, country=None):

        # To make queries you need to be logged in
        if not self.logged_in:  # pragma: no cover
            raise ConfigurationError("Cannot query if not logged in")

        # get the show id
        show_id = self.get_show_id(series, country)
        if show_id is None:
            logger.error("No show id found for %r ", series)
            return []

        # get the page of the season of the show
        logger.info("Getting the subtitle of show id %d, season %d episode %d, source %r", show_id, season, episode, source)
        subtitles = []

        # Default source is SDTV
        if not source or source.lower() == "hdtv":
            if resolution in ("1080i", "1080p", "720p"):
                sub_source = resolution
            else:
                sub_source = "normale"
        else:
            sub_source = source.lower()

        # Look for year
        params = {"apikey": self.apikey}
        r = self.session.get(self.server_url + "shows/" + str(show_id), params=params, timeout=30)
        r.raise_for_status()
        root = ElementTree.fromstring(r.content)

        year = root.find("data/show/started").text
        if year:
            year = int(year.split("-", 1)[0])
        tvdb_id = root.find("data/show/id_tvdb").text
        if tvdb_id:
            tvdb_id = int(tvdb_id)

        params = {"apikey": self.apikey, "show_id": show_id, "q": "%dx%02d" % (season, episode), "version": sub_source}
        r = self.session.get(self.server_url + "subtitles/search", params=params, timeout=30)
        r.raise_for_status()
        root = ElementTree.fromstring(r.content)

        if int(root.find("data/count").text) == 0:
            logger.warning("Subtitles not found,  try with rip suffix")

            params["version"] = sub_source + "rip"
            r = self.session.get(self.server_url + "subtitles/search", params=params, timeout=30)
            r.raise_for_status()
            root = ElementTree.fromstring(r.content)
            if int(root.find("data/count").text) == 0:
                logger.warning("Subtitles not found, go season mode")

                # If no subtitle are found for single episode try to download all season zip
                subs = self._get_season_subtitles(show_id, season, sub_source)
                if subs:
                    for subtitle in subs:
                        subtitle.source = source
                        subtitle.year = year
                        subtitle.tvdb_id = tvdb_id

                    return subs
                else:
                    return []

        # Looking for subtitles in first page
        for subtitle in root.findall("data/subtitles/subtitle"):
            if "%dx%02d" % (season, episode) in subtitle.find("name").text.lower():

                logger.debug("Found subtitle id %d - %r - %r", int(subtitle.find("id").text), subtitle.find("name").text, subtitle.find("version").text)

                sub = ItaSASubtitle(
                    int(subtitle.find("id").text), subtitle.find("show_name").text, season, episode, source, year, tvdb_id, subtitle.find("name").text
                )

                subtitles.append(sub)

        # Not in the first page of result try next (if any)
        next_page = root.find("data/next")
        while next_page.text is not None:  # pragma: no cover

            r = self.session.get(next_page.text, timeout=30)
            r.raise_for_status()
            root = ElementTree.fromstring(r.content)

            # logger.info('Loading subtitles page %r', root.data.page.text)

            # Looking for show in following pages
            for subtitle in root.findall("data/subtitles/subtitle"):
                if "%dx%02d" % (season, episode) in subtitle.find("name").text.lower():

                    logger.debug("Found subtitle id %d - %r - %r", int(subtitle.find("id").text), subtitle.find("name").text, subtitle.find("version").text)

                    sub = ItaSASubtitle(
                        int(subtitle.find("id").text), subtitle.find("show_name").text, season, episode, source, year, tvdb_id, subtitle.find("name").text
                    )

                    subtitles.append(sub)

            next_page = root.find("data/next")

        # Download the subs found, can be more than one in zip
        additional_subs = []
        for sub in subtitles:

            # open the zip
            content = self._download_zip(sub.sub_id)
            if not is_zipfile(io.BytesIO(content)):  # pragma: no cover
                if "limite di download" in content:
                    raise DownloadLimitExceeded()
                else:
                    raise ConfigurationError("Not a zip file: %r" % content)

            with ZipFile(io.BytesIO(content)) as zf:
                if len(zf.namelist()) > 1:  # pragma: no cover

                    for index, name in enumerate(zf.namelist()):

                        if index == 0:
                            # First element
                            sub.content = fix_line_ending(zf.read(name))
                            sub.full_data = name
                        else:
                            add_sub = copy.deepcopy(sub)
                            add_sub.content = fix_line_ending(zf.read(name))
                            add_sub.full_data = name
                            additional_subs.append(add_sub)
                else:
                    sub.content = fix_line_ending(zf.read(zf.namelist()[0]))
                    sub.full_data = zf.namelist()[0]

        return subtitles + additional_subs

    def list_subtitles(self, video: Episode, languages):
        return self.query(video.series, video.season, video.episode, video.source, video.resolution)

    def download_subtitle(self, subtitle):  # pragma: no cover
        pass



# -*- coding: utf-8 -*-

############################ Copyrights and license ############################
#                                                                              #
# Copyright 2013 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2014 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2016 Jannis Gebauer <ja.geb@me.com>                                #
# Copyright 2016 Peter Buckley <dx-pbuckley@users.noreply.github.com>          #
# Copyright 2018 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

import github.GithubObject


class Rate(github.GithubObject.NonCompletableGithubObject):
    """
    This class represents Rates. The reference can be found here http://developer.github.com/v3/rate_limit
    """

    def __repr__(self):
        return self.get__repr__(
            {
                "limit": self._limit.value,
                "remaining": self._remaining.value,
                "reset": self._reset.value,
            }
        )

    @property
    def limit(self):
        """
        :type: integer
        """
        return self._limit.value

    @property
    def remaining(self):
        """
        :type: integer
        """
        return self._remaining.value

    @property
    def reset(self):
        """
        :type: datetime.datetime
        """
        return self._reset.value

    def _initAttributes(self):
        self._limit = github.GithubObject.NotSet
        self._remaining = github.GithubObject.NotSet
        self._reset = github.GithubObject.NotSet

    def _useAttributes(self, attributes):
        if "limit" in attributes:  # pragma no branch
            self._limit = self._makeIntAttribute(attributes["limit"])
        if "remaining" in attributes:  # pragma no branch
            self._remaining = self._makeIntAttribute(attributes["remaining"])
        if "reset" in attributes:  # pragma no branch
            self._reset = self._makeTimestampAttribute(attributes["reset"])



import validators

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        # Provider Init
        super().__init__("TorrentProject")

        # Credentials
        self.public = True

        # Torrent Stats
        self.minseed = 0
        self.minleech = 0

        # URLs
        self.url = "https://torrentproject.to/"

        self.custom_url = None

        self.ability_status = self.PROVIDER_BACKLOG
        # Proper Strings

        # Cache
        self.cache = tvcache.TVCache(self, search_params={"RSS": ["0day"]})

    def search(self, search_strings, age=0, ep_obj=None):
        results = []

        search_params = {"out": "json", "filter": 2101, "showmagnets": "on", "num": 50}

        for mode in search_strings:  # Mode = RSS, Season, Episode
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))

            for search_string in {*search_strings[mode]}:

                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))

                search_params["s"] = search_string

                if self.custom_url:
                    if not validators.url(self.custom_url):
                        logger.warning("Invalid custom url set, please check your settings")
                        return results
                    search_url = self.custom_url
                else:
                    search_url = self.url

                torrents = self.get_url(search_url, params=search_params, returns="json")
                if not (torrents and "total_found" in torrents and int(torrents["total_found"]) > 0):
                    logger.debug("Data returned from provider does not contain any torrents")
                    continue

                del torrents["total_found"]

                results = []
                for i in torrents:
                    title = torrents[i]["title"]
                    seeders = try_int(torrents[i]["seeds"], 1)
                    leechers = try_int(torrents[i]["leechs"], 0)
                    if seeders < self.minseed or leechers < self.minleech:
                        if mode != "RSS":
                            logger.debug("Torrent doesn't meet minimum seeds & leechers not selecting : {0}".format(title))
                        continue

                    t_hash = torrents[i]["torrent_hash"]
                    torrent_size = torrents[i]["torrent_size"]
                    if not all([t_hash, torrent_size]):
                        continue
                    download_url = torrents[i]["magnet"] + self._custom_trackers
                    size = convert_size(torrent_size) or -1

                    if not all([title, download_url]):
                        continue

                    item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": t_hash}

                    if mode != "RSS":
                        logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                    items.append(item)

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)
            results += items

        return results



from __future__ import absolute_import

import os
import sys
import ctypes.util

from ctypes import c_void_p, c_size_t, byref, create_string_buffer, CDLL

from . import JavaScriptInterpreter
from .encapsulated import template

# ------------------------------------------------------------------------------- #


class ChallengeInterpreter(JavaScriptInterpreter):

    # ------------------------------------------------------------------------------- #

    def __init__(self):
        super(ChallengeInterpreter, self).__init__('chakracore')

    # ------------------------------------------------------------------------------- #

    def eval(self, body, domain):
        chakraCoreLibrary = None

        # check current working directory.
        for _libraryFile in ['libChakraCore.so', 'libChakraCore.dylib', 'ChakraCore.dll']:
            if os.path.isfile(os.path.join(os.getcwd(), _libraryFile)):
                chakraCoreLibrary = os.path.join(os.getcwd(), _libraryFile)
                continue

        if not chakraCoreLibrary:
            chakraCoreLibrary = ctypes.util.find_library('ChakraCore')

        if not chakraCoreLibrary:
            sys.tracebacklimit = 0
            raise RuntimeError(
                'ChakraCore library not found in current path or any of your system library paths, '
                'please download from https://www.github.com/VeNoMouS/cloudscraper/tree/ChakraCore/, '
                'or https://github.com/Microsoft/ChakraCore/'
            )

        try:
            chakraCore = CDLL(chakraCoreLibrary)
        except OSError:
            sys.tracebacklimit = 0
            raise RuntimeError('There was an error loading the ChakraCore library {}'.format(chakraCoreLibrary))

        if sys.platform != 'win32':
            chakraCore.DllMain(0, 1, 0)
            chakraCore.DllMain(0, 2, 0)

        script = create_string_buffer(template(body, domain).encode('utf-16'))

        runtime = c_void_p()
        chakraCore.JsCreateRuntime(0, 0, byref(runtime))

        context = c_void_p()
        chakraCore.JsCreateContext(runtime, byref(context))
        chakraCore.JsSetCurrentContext(context)

        fname = c_void_p()
        chakraCore.JsCreateString(
            'iuam-challenge.js',
            len('iuam-challenge.js'),
            byref(fname)
        )

        scriptSource = c_void_p()
        chakraCore.JsCreateExternalArrayBuffer(
            script,
            len(script),
            0,
            0,
            byref(scriptSource)
        )

        jsResult = c_void_p()
        chakraCore.JsRun(scriptSource, 0, fname, 0x02, byref(jsResult))

        resultJSString = c_void_p()
        chakraCore.JsConvertValueToString(jsResult, byref(resultJSString))

        stringLength = c_size_t()
        chakraCore.JsCopyString(resultJSString, 0, 0, byref(stringLength))

        resultSTR = create_string_buffer(stringLength.value + 1)
        chakraCore.JsCopyString(
            resultJSString,
            byref(resultSTR),
            stringLength.value + 1,
            0
        )

        chakraCore.JsDisposeRuntime(runtime)

        return resultSTR.value


# ------------------------------------------------------------------------------- #

ChallengeInterpreter()



import glob
import os
from abc import ABC

from sickchill import settings


class UpdateManagerBase(ABC):
    @staticmethod
    def get_update_url():
        return f"{settings.WEB_ROOT}/home/update/?pid={settings.PID}"

    @staticmethod
    def _clean_pyc(path):
        path_parts = [os.path.dirname(settings.PROG_DIR), path, "*.pyc"]
        for f in glob.iglob(os.path.join(*path_parts)):
            os.remove(f)

        path_parts.insert(-1, "**")
        for f in glob.iglob(os.path.join(*path_parts)):
            os.remove(f)

    def list_remote_branches(self):
        raise NotImplementedError

    def get_current_version(self):
        raise NotImplementedError

    def get_current_commit_hash(self):
        raise NotImplementedError

    def get_newest_version(self):
        raise NotImplementedError

    def get_newest_commit_hash(self):
        raise NotImplementedError

    def get_num_commits_behind(self):
        raise NotImplementedError

    def set_newest_text(self):
        raise NotImplementedError

    def need_update(self):
        raise NotImplementedError

    def update(self):
        raise NotImplementedError



# vim: set fileencoding=utf-8 :
import os
import re
import locale
import json
import ctypes
import sys
import warnings
from pkg_resources import get_distribution, DistributionNotFound
import xml.etree.ElementTree as ET

try:
    import pathlib
except ImportError:
    pathlib = None

if sys.version_info < (3,):
    import urlparse
else:
    import urllib.parse as urlparse

try:
    __version__ = get_distribution("pymediainfo").version
except DistributionNotFound:
    pass

class Track(object):
    """
    An object associated with a media file track.

    Each :class:`Track` attribute corresponds to attributes parsed from MediaInfo's output.
    All attributes are lower case. Attributes that are present several times such as Duration
    yield a second attribute starting with `other_` which is a list of all alternative attribute values.

    When a non-existing attribute is accessed, `None` is returned.

    Example:

    >>> t = mi.tracks[0]
    >>> t
    <Track track_id='None', track_type='General'>
    >>> t.duration
    3000
    >>> t.to_data()["other_duration"]
    ['3 s 0 ms', '3 s 0 ms', '3 s 0 ms',
        '00:00:03.000', '00:00:03.000']
    >>> type(t.non_existing)
    NoneType

    All available attributes can be obtained by calling :func:`to_data`.
    """
    def __eq__(self, other):
        return self.__dict__ == other.__dict__
    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except:
            pass
        return None
    def __getstate__(self):
        return self.__dict__
    def __setstate__(self, state):
        self.__dict__ = state
    def __init__(self, xml_dom_fragment):
        self.track_type = xml_dom_fragment.attrib['type']
        for el in xml_dom_fragment:
            node_name = el.tag.lower().strip().strip('_')
            if node_name == 'id':
                node_name = 'track_id'
            node_value = el.text
            other_node_name = "other_%s" % node_name
            if getattr(self, node_name) is None:
                setattr(self, node_name, node_value)
            else:
                if getattr(self, other_node_name) is None:
                    setattr(self, other_node_name, [node_value, ])
                else:
                    getattr(self, other_node_name).append(node_value)

        for o in [d for d in self.__dict__.keys() if d.startswith('other_')]:
            try:
                primary = o.replace('other_', '')
                setattr(self, primary, int(getattr(self, primary)))
            except:
                for v in getattr(self, o):
                    try:
                        current = getattr(self, primary)
                        setattr(self, primary, int(v))
                        getattr(self, o).append(current)
                        break
                    except:
                        pass
    def __repr__(self):
        return("<Track track_id='{}', track_type='{}'>".format(self.track_id, self.track_type))
    def to_data(self):
        """
        Returns a dict representation of the track attributes.

        Example:

        >>> sorted(track.to_data().keys())[:3]
        ['codec', 'codec_extensions_usually_used', 'codec_url']
        >>> t.to_data()["file_size"]
        5988


        :rtype: dict
        """
        data = {}
        for k, v in self.__dict__.items():
            if k != 'xml_dom_fragment':
                data[k] = v
        return data


class MediaInfo(object):
    """
    An object containing information about a media file.


    :class:`MediaInfo` objects can be created by directly calling code from
    libmediainfo (in this case, the library must be present on the system):

    >>> pymediainfo.MediaInfo.parse("/path/to/file.mp4")

    Alternatively, objects may be created from MediaInfo's XML output.
    Such output can be obtained using the ``XML`` output format on versions older than v17.10
    and the ``OLDXML`` format on newer versions.

    Using such an XML file, we can create a :class:`MediaInfo` object:

    >>> with open("output.xml") as f:
    ...     mi = pymediainfo.MediaInfo(f.read())

    :param str xml: XML output obtained from MediaInfo.
    :param str encoding_errors: option to pass to :func:`str.encode`'s `errors`
        parameter before parsing `xml`.
    :raises xml.etree.ElementTree.ParseError: if passed invalid XML.
    :var tracks: A list of :py:class:`Track` objects which the media file contains.
        For instance:

        >>> mi = pymediainfo.MediaInfo.parse("/path/to/file.mp4")
        >>> for t in mi.tracks:
        ...     print(t)
        <Track track_id='None', track_type='General'>
        <Track track_id='1', track_type='Text'>
    """
    def __eq__(self, other):
        return self.tracks == other.tracks
    def __init__(self, xml, encoding_errors="strict"):
        xml_dom = ET.fromstring(xml.encode("utf-8", encoding_errors))
        self.tracks = []
        # This is the case for libmediainfo < 18.03
        # https://github.com/sbraz/pymediainfo/issues/57
        # https://github.com/MediaArea/MediaInfoLib/commit/575a9a32e6960ea34adb3bc982c64edfa06e95eb
        if xml_dom.tag == "File":
            xpath = "track"
        else:
            xpath = "File/track"
        for xml_track in xml_dom.iterfind(xpath):
            self.tracks.append(Track(xml_track))
    @staticmethod
    def _parse_filename(filename):
        if hasattr(os, "PathLike") and isinstance(filename, os.PathLike):
            return os.fspath(filename), False
        elif pathlib is not None and isinstance(filename, pathlib.PurePath):
            return str(filename), False
        else:
            url = urlparse.urlparse(filename)
            return filename, bool(url.scheme)
    @staticmethod
    def _get_library(library_file=None):
        os_is_nt = os.name in ("nt", "dos", "os2", "ce")
        if os_is_nt:
            lib_type = ctypes.WinDLL
        else:
            lib_type = ctypes.CDLL
        if library_file is None:
            if os_is_nt:
                library_names = ("MediaInfo.dll",)
            elif sys.platform == "darwin":
                library_names = ("libmediainfo.0.dylib", "libmediainfo.dylib")
            else:
                library_names = ("libmediainfo.so.0",)
            script_dir = os.path.dirname(__file__)
            # Look for the library file in the script folder
            for library in library_names:
                lib_path = os.path.join(script_dir, library)
                if os.path.isfile(lib_path):
                    # If we find it, don't try any other filename
                    library_names = (lib_path,)
                    break
        else:
            library_names = (library_file,)
        for i, library in enumerate(library_names, start=1):
            try:
                lib = lib_type(library)
                # Define arguments and return types
                lib.MediaInfo_Inform.restype = ctypes.c_wchar_p
                lib.MediaInfo_New.argtypes = []
                lib.MediaInfo_New.restype  = ctypes.c_void_p
                lib.MediaInfo_Option.argtypes = [ctypes.c_void_p, ctypes.c_wchar_p, ctypes.c_wchar_p]
                lib.MediaInfo_Option.restype = ctypes.c_wchar_p
                lib.MediaInfo_Inform.argtypes = [ctypes.c_void_p, ctypes.c_size_t]
                lib.MediaInfo_Inform.restype = ctypes.c_wchar_p
                lib.MediaInfo_Open.argtypes = [ctypes.c_void_p, ctypes.c_wchar_p]
                lib.MediaInfo_Open.restype = ctypes.c_size_t
                lib.MediaInfo_Delete.argtypes = [ctypes.c_void_p]
                lib.MediaInfo_Delete.restype  = None
                lib.MediaInfo_Close.argtypes = [ctypes.c_void_p]
                lib.MediaInfo_Close.restype = None
                # Without a handle, there might be problems when using concurrent threads
                # https://github.com/sbraz/pymediainfo/issues/76#issuecomment-574759621
                handle = lib.MediaInfo_New()
                lib_version_str = lib.MediaInfo_Option(handle, "Info_Version", "")
                lib_version_str = re.search(r"^MediaInfoLib - v(\S+)", lib_version_str).group(1)
                lib_version = tuple(int(_) for _ in lib_version_str.split("."))
                return (lib, handle, lib_version_str, lib_version)
            except OSError:
                # If we've tried all possible filenames
                if i == len(library_names):
                    raise
    @classmethod
    def can_parse(cls, library_file=None):
        """
        Checks whether media files can be analyzed using libmediainfo.

        :rtype: bool
        """
        try:
            lib, handle = cls._get_library(library_file)[:2]
            lib.MediaInfo_Close(handle)
            lib.MediaInfo_Delete(handle)
            return True
        except:
            return False
    @classmethod
    def parse(cls, filename, library_file=None, cover_data=False,
            encoding_errors="strict", parse_speed=0.5, text=False,
            full=True, legacy_stream_display=False, mediainfo_options=None,
            output=None):
        """
        Analyze a media file using libmediainfo.

        .. note::
            Because of the way the underlying library works, this method should not
            be called simultaneously from multiple threads *with different arguments*.
            Doing so will cause inconsistencies or failures by changing
            library options that are shared across threads.

        :param filename: path to the media file which will be analyzed.
            A URL can also be used if libmediainfo was compiled
            with CURL support.
        :param str library_file: path to the libmediainfo library, this should only be used if the library cannot be auto-detected.
        :param bool cover_data: whether to retrieve cover data as base64.
        :param str encoding_errors: option to pass to :func:`str.encode`'s `errors`
            parameter before parsing MediaInfo's XML output.
        :param float parse_speed: passed to the library as `ParseSpeed`,
            this option takes values between 0 and 1.
            A higher value will yield more precise results in some cases
            but will also increase parsing time.
        :param bool full: display additional tags, including computer-readable values
            for sizes and durations.
        :param bool legacy_stream_display: display additional information about streams.
        :param dict mediainfo_options: additional options that will be passed to the `MediaInfo_Option` function,
            for example: ``{"Language": "raw"}``. Do not use this parameter when running the
            method simultaneously from multiple threads, it will trigger a reset of all options
            which will cause inconsistencies or failures.
        :param str output: custom output format for MediaInfo, corresponds to the CLI's
            ``--Output`` parameter. Setting this causes the method to
            return a `str` instead of a :class:`MediaInfo` object.

            Useful values include:
                * the empty `str` ``""`` (corresponds to the default
                  text output, obtained when running ``mediainfo`` with no
                  additional parameters)

                * ``"XML"``

                * ``"JSON"``

                * ``%``-delimited templates (see ``mediainfo --Info-Parameters``)
        :type filename: str or pathlib.Path or os.PathLike
        :rtype: str if `output` is set.
        :rtype: :class:`MediaInfo` otherwise.
        :raises FileNotFoundError: if passed a non-existent file
            (Python ≥ 3.3), does not work on Windows.
        :raises IOError: if passed a non-existent file (Python < 3.3),
            does not work on Windows.
        :raises RuntimeError: if parsing fails, this should not
            happen unless libmediainfo itself fails.

        Examples:
            >>> pymediainfo.MediaInfo.parse("tests/data/sample.mkv")
                <pymediainfo.MediaInfo object at 0x7fa83a3db240>

            >>> import json
            >>> mi = pymediainfo.MediaInfo.parse("tests/data/sample.mkv",
            ...     output="JSON")
            >>> json.loads(mi)["media"]["track"][0]
                {'@type': 'General', 'TextCount': '1', 'FileExtension': 'mkv',
                    'FileSize': '5904',  … }


        """
        lib, handle, lib_version_str, lib_version = cls._get_library(library_file)
        filename, is_url = cls._parse_filename(filename)
        # Try to open the file (if it's not a URL)
        # Doesn't work on Windows because paths are URLs
        if not is_url:
            # Test whether the file is readable
            with open(filename, "rb"):
                pass
        # The XML option was renamed starting with version 17.10
        if lib_version >= (17, 10):
            xml_option = "OLDXML"
        else:
            xml_option = "XML"
        # Cover_Data is not extracted by default since version 18.03
        # See https://github.com/MediaArea/MediaInfoLib/commit/d8fd88a1c282d1c09388c55ee0b46029e7330690
        if lib_version >= (18, 3):
            lib.MediaInfo_Option(handle, "Cover_Data", "base64" if cover_data else "")
        lib.MediaInfo_Option(handle, "CharSet", "UTF-8")
        # Fix for https://github.com/sbraz/pymediainfo/issues/22
        # Python 2 does not change LC_CTYPE
        # at startup: https://bugs.python.org/issue6203
        if (sys.version_info < (3,) and os.name == "posix"
                and locale.getlocale() == (None, None)):
            locale.setlocale(locale.LC_CTYPE, locale.getdefaultlocale())
        if text:
            warnings.warn('The "text" option is obsolete and will be removed '
                    'in the next major version. Use output="" instead.',
                    DeprecationWarning
            )
            output = ""
        lib.MediaInfo_Option(handle, "Inform", xml_option if output is None else output)
        lib.MediaInfo_Option(handle, "Complete", "1" if full else "")
        lib.MediaInfo_Option(handle, "ParseSpeed", str(parse_speed))
        lib.MediaInfo_Option(handle, "LegacyStreamDisplay", "1" if legacy_stream_display else "")
        if mediainfo_options is not None:
            if lib_version < (19, 9):
                warnings.warn("This version of MediaInfo ({}) does not support resetting all options to their default values, "
                    "passing it custom options is not recommended and may result in unpredictable behavior, "
                    "see https://github.com/MediaArea/MediaInfoLib/issues/1128".format(lib_version_str),
                    RuntimeWarning
                )
            for option_name, option_value in mediainfo_options.items():
                lib.MediaInfo_Option(handle, option_name, option_value)
        if lib.MediaInfo_Open(handle, filename) == 0:
            lib.MediaInfo_Close(handle)
            lib.MediaInfo_Delete(handle)
            raise RuntimeError("An eror occured while opening {}"
                    " with libmediainfo".format(filename))
        info = lib.MediaInfo_Inform(handle, 0)
        # Reset all options to their defaults so that they aren't
        # retained when the parse method is called several times
        # https://github.com/MediaArea/MediaInfoLib/issues/1128
        # Do not call it when it is not required because it breaks threads
        # https://github.com/sbraz/pymediainfo/issues/76#issuecomment-575245093
        if mediainfo_options is not None and lib_version >= (19, 9):
            lib.MediaInfo_Option(handle, "Reset", "")
        # Delete the handle
        lib.MediaInfo_Close(handle)
        lib.MediaInfo_Delete(handle)
        if output is None:
            return cls(info, encoding_errors)
        else:
            return info
    def to_data(self):
        """
        Returns a dict representation of the object's :py:class:`Tracks <Track>`.

        :rtype: dict
        """
        data = {'tracks': []}
        for track in self.tracks:
            data['tracks'].append(track.to_data())
        return data
    def to_json(self):
        """
        Returns a JSON representation of the object's :py:class:`Tracks <Track>`.

        :rtype: str
        """
        return json.dumps(self.to_data())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2016-5-26
@author: thinhlx
'''

LOCALE = [
    ["gerade eben", "in einer Weile"],
    ["vor %s Sekunden", "in %s Sekunden"],
    ["vor 1 Minute", "in 1 Minute"],
    ["vor %s Minuten", "in %s Minuten"],
    ["vor einer Stunde", "in 1 Stunde"],
    ["vor %s Stunden", "in %s Stunden"],
    ["vor 1 Tag", "in 1 Tag"],
    ["vor %s Tagen", "in %s Tagen"],
    ["vor 1 Woche", "in 1 Woche"],
    ["vor %s Wochen", "in %s Wochen"],
    ["vor 1 Monat", "in 1 Monat"],
    ["vor %s Monaten", "in %s Monaten"],
    ["vor 1 Jahr", "in 1 Jahr"],
    ["vor %s Jahren", "in %s Jahren"],
]
